#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import functools

import luigi

from servicecatalog_puppet import aws, constants
from servicecatalog_puppet.workflow.dependencies import tasks


class TerminateStackTask(tasks.TaskWithReference):
    stack_name = luigi.Parameter()

    region = luigi.Parameter()
    account_id = luigi.Parameter()

    bucket = luigi.Parameter()
    key = luigi.Parameter()
    version_id = luigi.Parameter()

    launch_name = luigi.Parameter()
    stack_set_name = luigi.Parameter()
    capabilities = luigi.ListParameter()

    use_service_role = luigi.BoolParameter()

    ssm_param_inputs = luigi.ListParameter(default=[], significant=False)

    launch_parameters = luigi.DictParameter(default={}, significant=False)
    manifest_parameters = luigi.DictParameter(default={}, significant=False)
    account_parameters = luigi.DictParameter(default={}, significant=False)

    retry_count = luigi.IntParameter(default=1, significant=False)
    worker_timeout = luigi.IntParameter(default=0, significant=False)
    ssm_param_outputs = luigi.ListParameter(default=[], significant=False)
    requested_priority = luigi.IntParameter(significant=False, default=0)

    execution = luigi.Parameter()

    try_count = 1
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "stack_name": self.stack_name,
            "account_id": self.account_id,
            "region": self.region,
        }

    @property
    @functools.lru_cache(maxsize=32)
    def stack_name_to_use(self):
        if self.launch_name != "":
            with self.spoke_regional_client("servicecatalog") as servicecatalog:
                try:
                    pp_id = (
                        servicecatalog.describe_provisioned_product(
                            Name=self.launch_name
                        )
                        .get("ProvisionedProductDetail")
                        .get("Id")
                    )
                except servicecatalog.exceptions.ResourceNotFoundException as e:
                    if (
                        "Provisioned product not found"
                        in e.response["Error"]["Message"]
                    ):
                        return self.stack_name
                    else:
                        raise e
                stack_name = aws.get_stack_name_for_pp_id(servicecatalog, pp_id)
                return stack_name

        elif self.stack_set_name != "":
            with self.spoke_regional_client("cloudformation") as cloudformation:
                paginator = cloudformation.get_paginator("list_stacks")
                for page in paginator.paginate():
                    for summary in page.get("StackSummaries", []):
                        if summary.get("StackName").startswith(
                            f"StackSet-{self.stack_set_name}-"
                        ):
                            return summary.get("StackName")
                raise Exception(
                    f"Could not find a stack beginning with StackSet-{self.stack_set_name}- in {self.region} of {self.account_id}"
                )

        return self.stack_name

    def terminate_ssm_outputs(self):
        outputs = (
            self.get_from_manifest(constants.STACKS, self.stack_name)
            .get("outputs", {})
            .get("ssm", [])
        )
        for ssm_param_output in outputs:
            param_name = ssm_param_output.get("param_name")
            param_name = param_name.replace("${AWS::Region}", self.region)
            param_name = param_name.replace("${AWS::AccountId}", self.account_id)
            self.info(f"deleting SSM Param: {param_name}")
            with self.hub_client("ssm") as ssm:
                try:
                    ssm.delete_parameter(
                        Name=param_name,
                    )
                    self.info(f"deleting SSM Param: {param_name}")
                except ssm.exceptions.ParameterNotFound:
                    self.info(f"SSM Param: {param_name} not found")

    def run(self):
        if self.execution == constants.EXECUTION_MODE_HUB:
            self.terminate_ssm_outputs()

        with self.spoke_regional_client("cloudformation") as cloudformation:
            cloudformation.ensure_deleted(StackName=self.stack_name_to_use)

        self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import json
import os
from datetime import datetime
from pathlib import Path

from servicecatalog_puppet import constants


def record_event(event_type, task, extra_event_data=None):
    task_type = task.__class__.__name__
    task_params = task.param_kwargs
    current_datetime = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    event = {
        "event_type": event_type,
        "task_type": task_type,
        "task_params": task_params,
        "params_for_results": task.params_for_results_display(),
        "datetime": current_datetime,
        "pid": os.getpid(),
    }
    if extra_event_data is not None:
        event.update(extra_event_data)

    with open(
        Path(constants.RESULTS_DIRECTORY)
        / event_type
        / f"{task_type}-{task.task_id}.json",
        "w",
    ) as f:
        f.write(
            json.dumps(
                event,
                default=str,
                indent=4,
            )
        )



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

COMPLETED = "COMPLETED"
NOT_SET = "NOT_SET"
ERRORED = "ERRORED"
QUEUE_STATUS = "QUEUE_STATUS"
IN_PROGRESS = "IN_PROGRESS"
BLOCKED = "BLOCKED"
RESOURCES_REQUIRED = "resources_required"

CONTROL_EVENT__COMPLETE = "CONTROL_EVENT__COMPLETE"



#   Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#   SPDX-License-Identifier: Apache-2.0


MANIFEST_SECTION_NAMES = "manifest_section_names"
MANIFEST_ITEM_NAMES = "manifest_item_names"
MANIFEST_ACCOUNT_IDS = "manifest_account_ids"



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import functools
import re
import time

import cfn_tools
import luigi
from botocore.exceptions import ClientError

from servicecatalog_puppet import aws, config, constants
from servicecatalog_puppet.workflow.dependencies import tasks


class ProvisionStackTask(tasks.TaskWithParameters):
    stack_name = luigi.Parameter()

    region = luigi.Parameter()
    account_id = luigi.Parameter()

    bucket = luigi.Parameter()
    key = luigi.Parameter()
    version_id = luigi.Parameter()

    launch_name = luigi.Parameter()
    stack_set_name = luigi.Parameter()
    get_s3_template_ref = luigi.Parameter()
    capabilities = luigi.ListParameter()

    ssm_param_inputs = luigi.ListParameter(default=[], significant=False)

    launch_parameters = luigi.DictParameter(default={}, significant=False)
    manifest_parameters = luigi.DictParameter(default={}, significant=False)
    account_parameters = luigi.DictParameter(default={}, significant=False)

    retry_count = luigi.IntParameter(default=1, significant=False)
    worker_timeout = luigi.IntParameter(default=0, significant=False)
    ssm_param_outputs = luigi.ListParameter(default=[], significant=False)
    requested_priority = luigi.IntParameter(significant=False, default=0)

    use_service_role = luigi.BoolParameter()

    execution = luigi.Parameter()
    manifest_file_path = luigi.Parameter()

    tags = luigi.ListParameter()

    section_name = constants.STACKS
    cachable_level = constants.CACHE_LEVEL_TASK

    need_to_provision = False

    @property
    def item_name(self):
        return self.stack_name

    try_count = 1

    @property
    def drift_token_parameters(self):
        result = (
            f"stack_name/{self.stack_name}/"
            f"launch_name/{self.launch_name}/"
            f"stack_set_name/{self.stack_set_name}/"
            f"bucket/{self.bucket}/"
            f"key/{self.key}/"
            f"execution/{self.execution}"
        )
        return result

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "stack_name": self.stack_name,
            "account_id": self.account_id,
            "region": self.region,
            "need_to_provision": self.need_to_provision,
        }

    @property
    def priority(self):
        return self.requested_priority

    @property
    @functools.lru_cache(maxsize=32)
    def stack_name_to_use(self):
        if self.launch_name != "":
            with self.spoke_regional_client("servicecatalog") as servicecatalog:
                if "*" in self.launch_name:
                    paginator = servicecatalog.get_paginator(
                        "scan_provisioned_products"
                    )
                    for page in paginator.paginate(
                        AccessLevelFilter={"Key": "Account", "Value": "self"},
                    ):
                        for provisioned_product in page.get("ProvisionedProducts", []):
                            name_as_a_regex = self.launch_name.replace("*", "(.*)")
                            if re.match(
                                name_as_a_regex, provisioned_product.get("Name")
                            ):
                                pp_stack_name = aws.get_stack_name_for_pp_id(
                                    servicecatalog, provisioned_product.get("Id")
                                )
                                return pp_stack_name

                    return self.stack_name
                else:
                    try:
                        pp_id = (
                            servicecatalog.describe_provisioned_product(
                                Name=self.launch_name
                            )
                            .get("ProvisionedProductDetail")
                            .get("Id")
                        )
                    except servicecatalog.exceptions.ResourceNotFoundException as e:
                        if (
                            "Provisioned product not found"
                            in e.response["Error"]["Message"]
                        ):
                            return self.stack_name
                        else:
                            raise e
                    pp_stack_name = aws.get_stack_name_for_pp_id(servicecatalog, pp_id)
                    return pp_stack_name

        elif self.stack_set_name != "":
            with self.spoke_regional_client("cloudformation") as cloudformation:
                paginator = cloudformation.get_paginator("list_stacks")
                for page in paginator.paginate():
                    for summary in page.get("StackSummaries", []):
                        if summary.get("StackName").startswith(
                            f"StackSet-{self.stack_set_name}-"
                        ):
                            return summary.get("StackName")
                raise Exception(
                    f"Could not find a stack beginning with StackSet-{self.stack_set_name}- in {self.region} of {self.account_id}"
                )

        return self.stack_name

    def ensure_stack_is_in_complete_status(self):
        waiting = "NotARealStatus"
        current_stack = dict(StackStatus=waiting)
        with self.spoke_regional_client("cloudformation") as cloudformation:
            while current_stack.get(
                "StackStatus"
            ) in constants.CLOUDFORMATION_IN_PROGRESS_STATUS + [waiting]:
                try:
                    stacks = cloudformation.describe_stacks(
                        StackName=self.stack_name_to_use
                    ).get("Stacks", [])
                    assert len(stacks) == 1
                    current_stack = stacks[0]

                    if (
                        current_stack.get("StackStatus")
                        in constants.CLOUDFORMATION_IN_PROGRESS_STATUS
                    ):
                        time.sleep(2)
                except ClientError as error:
                    if (
                        error.response["Error"]["Message"]
                        == f"Stack with id {self.stack_name} does not exist"
                    ):
                        return dict(StackStatus="NoStack")
                    else:
                        raise error

        if current_stack.get("StackStatus") in constants.CLOUDFORMATION_UNHAPPY_STATUS:
            raise Exception(
                f"stack {self.stack_name} is in state {current_stack.get('StackStatus')}"
            )

        return current_stack

    def run(self):
        stack = self.ensure_stack_is_in_complete_status()
        status = stack.get("StackStatus")

        if status != "NoStack":
            with self.spoke_regional_client("cloudformation") as cloudformation:
                if status == "ROLLBACK_COMPLETE":
                    if self.should_delete_rollback_complete_stacks:
                        cloudformation.ensure_deleted(StackName=self.stack_name_to_use)
                    else:
                        raise Exception(
                            f"Stack: {self.stack_name_to_use} is in ROLLBACK_COMPLETE and need remediation"
                        )

        task_output = dict(
            **self.params_for_results_display(),
            stack_name_used=self.stack_name_to_use,
        )

        all_params = self.get_parameter_values()

        template_to_provision_source = self.get_output_from_reference_dependency_raw(
            self.get_s3_template_ref
        )

        try:
            template_to_provision = cfn_tools.load_yaml(template_to_provision_source)
        except Exception:
            try:
                template_to_provision = cfn_tools.load_json(
                    template_to_provision_source
                )
            except Exception:
                raise Exception("Could not parse new template as YAML or JSON")

        params_to_use = dict()
        for param_name, p in template_to_provision.get("Parameters", {}).items():
            if all_params.get(param_name, p.get("DefaultValue")) is not None:
                params_to_use[param_name] = all_params.get(
                    param_name, p.get("DefaultValue")
                )

        existing_stack_params_dict = dict()
        existing_template_from_cloudformation = ""
        if status != "NoStack":
            if status in constants.CLOUDFORMATION_HAPPY_STATUS:
                with self.spoke_regional_client("cloudformation") as cloudformation:
                    summary_response = cloudformation.get_template_summary(
                        StackName=self.stack_name_to_use,
                    )
                    for parameter in summary_response.get("Parameters", []):
                        existing_stack_params_dict[parameter.get("ParameterKey")] = (
                            parameter.get("DefaultValue")
                        )
                    for stack_param in stack.get("Parameters", []):
                        existing_stack_params_dict[stack_param.get("ParameterKey")] = (
                            stack_param.get("ParameterValue")
                        )
                    template_body = cloudformation.get_template(
                        StackName=self.stack_name_to_use, TemplateStage="Original"
                    ).get("TemplateBody")
                    if isinstance(template_body, dict):
                        existing_template_from_cloudformation = template_body
                    else:
                        try:
                            existing_template_from_cloudformation = cfn_tools.load_yaml(
                                template_body
                            )
                        except Exception:
                            try:
                                existing_template_from_cloudformation = (
                                    cfn_tools.load_json(template_body)
                                )
                            except Exception:
                                raise Exception(
                                    "Could not parse existing template as YAML or JSON"
                                )

        if status in ["UPDATE_ROLLBACK_COMPLETE", "NoStack"]:
            self.need_to_provision = True
        else:
            if existing_stack_params_dict == params_to_use:
                self.info(f"params unchanged")
                if cfn_tools.dump_yaml(template_to_provision) == cfn_tools.dump_yaml(
                    existing_template_from_cloudformation
                ):
                    self.info(f"template the same")
                    self.need_to_provision = False
                else:
                    self.info(f"template changed")
                    self.need_to_provision = True
            else:
                self.info(f"params changed")
                self.need_to_provision = True

        if self.need_to_provision:
            provisioning_parameters = []
            for p in params_to_use.keys():
                provisioning_parameters.append(
                    {"ParameterKey": p, "ParameterValue": params_to_use.get(p)}
                )
            with self.spoke_regional_client("cloudformation") as cloudformation:
                a = dict(
                    StackName=self.stack_name_to_use,
                    TemplateBody=template_to_provision_source.decode("utf-8"),
                    ShouldUseChangeSets=False,
                    Capabilities=self.capabilities,
                    Parameters=provisioning_parameters,
                    ShouldDeleteRollbackComplete=self.should_delete_rollback_complete_stacks,
                    Tags=self.initialiser_stack_tags,
                )
                if self.use_service_role:
                    a["RoleARN"] = config.get_puppet_stack_role_arn(self.account_id)
                if self.tags:
                    a["Tags"] = [
                        dict(Key=t.get("key"), Value=t.get("value")) for t in self.tags
                    ]
                cloudformation.create_or_update(**a)

        task_output["provisioned"] = self.need_to_provision
        task_output["section_name"] = self.section_name
        self.write_output(task_output)



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks


class DisassociateProductsFromPortfolio(tasks.TaskWithReference):
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    portfolio = luigi.Parameter()
    portfolio_task_reference = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "account_id": self.account_id,
            "region": self.region,
            "portfolio": self.portfolio,
        }

    def run(self):
        portfolio = self.get_output_from_reference_dependency(
            self.portfolio_task_reference
        )
        portfolio_id = portfolio.get("Id")

        if portfolio_id is None:
            self.write_empty_output()
            return
        else:
            disassociations = list()
            with self.spoke_regional_client("servicecatalog") as servicecatalog:
                paginator = servicecatalog.get_paginator("search_products_as_admin")
                for page in paginator.paginate(PortfolioId=portfolio_id):
                    for product_view_details in page.get("ProductViewDetails", []):
                        product_id = product_view_details.get(
                            "ProductViewSummary", {}
                        ).get("ProductId")
                        servicecatalog.disassociate_product_from_portfolio(
                            PortfolioId=portfolio_id,
                            ProductId=product_id,
                        )
                        disassociations.append(
                            dict(portfolio_id=portfolio_id, product_id=product_id)
                        )

            self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from unittest import skip

from servicecatalog_puppet.workflow import tasks_unit_tests_helper


class GetCloudFormationTemplateFromS3Test(tasks_unit_tests_helper.PuppetTaskUnitTest):
    account_id = "account_id"
    bucket = "bucket"
    key = "key"
    region = "region"
    version_id = "version_id"

    def setUp(self) -> None:
        from servicecatalog_puppet.workflow.stack import (
            get_cloud_formation_template_from_s3,
        )

        self.module = get_cloud_formation_template_from_s3

        self.sut = self.module.GetCloudFormationTemplateFromS3(
            **self.get_common_args(),
            account_id=self.account_id,
            bucket=self.bucket,
            key=self.key,
            region=self.region,
            version_id=self.version_id,
        )

        self.wire_up_mocks()

    def test_params_for_results_display(self):
        # setup
        expected_result = {
            "task_reference": self.task_reference,
            "bucket": self.bucket,
            "key": self.key.replace("-${AWS::Region}", f"-{self.region}"),
            "region": self.region,
            "version_id": self.version_id,
        }

        # exercise
        actual_result = self.sut.params_for_results_display()

        # verify
        self.assertEqual(expected_result, actual_result)

    @skip
    def test_run(self):
        # setup
        # exercise
        actual_result = self.sut.run()

        # verify
        raise NotImplementedError()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import deepdiff
import jmespath
import luigi
from deepmerge import always_merger

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks


class DoAssertTask(tasks.TaskWithParameters):
    assertion_name = luigi.Parameter()
    region = luigi.Parameter()
    account_id = luigi.Parameter()

    execution = luigi.Parameter()

    expected = luigi.DictParameter()
    actual = luigi.DictParameter()

    requested_priority = luigi.IntParameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "assertion_name": self.assertion_name,
            "region": self.region,
            "account_id": self.account_id,
        }

    def get_actual_result(self):
        config = self.actual.get("config")
        client = config.get("client")
        use_paginator = config.get("use_paginator")
        call = config.get("call")
        arguments = config.get("arguments")
        filter = config.get("filter")

        with self.spoke_regional_client(client) as client:
            if use_paginator:
                paginator = client.get_paginator(call)
                result = dict()
                for page in paginator.paginate(**arguments):
                    always_merger.merge(result, page)
            else:
                f = getattr(client, call)
                result = f(**arguments)

        actual_result = jmespath.search(filter, result)
        if isinstance(actual_result, str):
            return actual_result.strip()
        else:
            return actual_result

    def run(self):
        actual_result = self.get_actual_result()
        expected_result = self.expected.get("config").get("value")
        if isinstance(expected_result, tuple):
            expected_result = list(expected_result)
        elif isinstance(actual_result, str):
            expected_result = expected_result.strip()

        ddiff = deepdiff.DeepDiff(actual_result, expected_result, ignore_order=True)

        if len(ddiff.keys()) > 0:
            raise Exception(ddiff)
        else:
            self.write_empty_output()



#   Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#   SPDX-License-Identifier: Apache-2.0
from servicecatalog_puppet import constants, task_reference_constants


def s3_parameter_handler(
    all_tasks, home_region, new_tasks, parameter_details, puppet_account_id, task
):
    if parameter_details.get("s3"):
        s3_parameter_details = parameter_details.get("s3")
        key = str(s3_parameter_details.get("key"))
        jmespath = str(s3_parameter_details.get("jmespath"))
        default = str(s3_parameter_details.get("default"))

        task_account_id = task.get("account_id")
        task_region = task.get("region")

        key = (
            key.replace("${AWS::Region}", task_region)
            .replace("${AWS::AccountId}", task_account_id)
            .replace("${AWS::PuppetAccountId}", puppet_account_id)
        )
        jmespath = (
            jmespath.replace("${AWS::Region}", task_region)
            .replace("${AWS::AccountId}", task_account_id)
            .replace("${AWS::PuppetAccountId}", puppet_account_id)
        )
        default = (
            default.replace("${AWS::Region}", task_region)
            .replace("${AWS::AccountId}", task_account_id)
            .replace("${AWS::PuppetAccountId}", puppet_account_id)
        )

        s3_object_task_reference = f"{constants.S3_GET_OBJECT}-{key}"

        if not all_tasks.get(s3_object_task_reference):
            new_tasks[s3_object_task_reference] = {
                "task_reference": s3_object_task_reference,
                "account_id": puppet_account_id,
                "region": home_region,
                "key": key,
                task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
                "dependencies": [],
                "dependencies_by_reference": [],
                "execution": constants.EXECUTION_MODE_HUB,
                "section_name": constants.S3_GET_OBJECT,
            }

        new_tasks[s3_object_task_reference][
            task_reference_constants.MANIFEST_SECTION_NAMES
        ].update(**task.get(task_reference_constants.MANIFEST_SECTION_NAMES))
        new_tasks[s3_object_task_reference][
            task_reference_constants.MANIFEST_ITEM_NAMES
        ].update(**task.get(task_reference_constants.MANIFEST_ITEM_NAMES))
        new_tasks[s3_object_task_reference][
            task_reference_constants.MANIFEST_ACCOUNT_IDS
        ].update(**task.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))

        task["dependencies_by_reference"].append(s3_object_task_reference)

        if not task.get("s3_object_task_reference"):
            task["s3_object_task_reference"] = dict()

        parameter_name = (
            str(s3_parameter_details.get("name"))
            .replace("${AWS::Region}", task_region)
            .replace("${AWS::AccountId}", task_account_id)
            .replace("${AWS::PuppetAccountId}", puppet_account_id)
        )

        task["s3_object_task_reference"][parameter_name] = s3_object_task_reference



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import datetime
import json

import click
import terminaltables
from betterboto import client as betterboto_client

from servicecatalog_puppet import constants, utils


result_look_up = dict(Failed="red", Succeeded="green")
trigger_look_up = dict(StartPipelineExecution="Manual Full Run")


def show_pipelines(format):
    pipelines_to_check = dict(
        core={"servicecatalog-puppet-pipeline": dict()},
    )

    fake_date = datetime.datetime.now()
    with betterboto_client.ClientContextManager("codepipeline") as codepipeline:
        for type, pipelines in pipelines_to_check.items():
            for pipeline_name, pipeline_details in pipelines.items():
                try:
                    executions = codepipeline.list_pipeline_executions(
                        pipelineName=pipeline_name, maxResults=5
                    ).get("pipelineExecutionSummaries", [])
                except codepipeline.exceptions.PipelineNotFoundException as e:
                    executions = []

                if len(executions) > 0:
                    last_execution = executions[0]
                else:
                    last_execution = dict(
                        status="N/A",
                        lastUpdateTime=fake_date,
                        startTime=fake_date,
                        sourceRevisions=[dict(revisionId="N/A", revisionSummary="N/A")],
                    )

                trend = list()
                for execution in executions:
                    trend.append(
                        dict(
                            start_time=execution.get("startTime", fake_date),
                            duration=execution.get("lastUpdateTime", fake_date)
                            - execution.get("startTime", fake_date),
                            status=execution.get("status"),
                            trigger=trigger_look_up.get(
                                execution.get("trigger").get("triggerType"), "N/A"
                            ),
                        )
                    )

                source_revisions = last_execution.get("sourceRevisions")
                if len(source_revisions) == 0:
                    source_revisions.append(
                        dict(revisionId="N/A", revisionSummary="N/A")
                    )

                pipelines[pipeline_name] = {
                    "name": pipeline_name,
                    "trigger": trigger_look_up.get(
                        last_execution.get("trigger").get("triggerType"), "N/A"
                    ),
                    "pipeline_execution_id": last_execution.get(
                        "pipelineExecutionId", "N/A"
                    ),
                    "start_time": last_execution.get("startTime", fake_date),
                    "status": last_execution.get("status"),
                    "revision_id": source_revisions[0].get("revisionId", "N/A"),
                    "revision_summary": source_revisions[0].get(
                        "revisionSummary", "N/A"
                    ),
                    "duration": last_execution.get("lastUpdateTime")
                    - last_execution.get("startTime"),
                    "trend": trend,
                }

    if format == "json":
        click.echo(json.dumps(pipelines_to_check, indent=4, default=str))
    elif format == "table":
        table_data = [
            [
                "Type",
                "Name",
                "Trigger",
                "Execution Id",
                "Start Time",
                "Status",
                "Last Commit Id",
                "Last Commit Message",
                "Duration",
                "Trend",
            ],
        ]
        for type, pipelines in pipelines_to_check.items():
            for pipeline_name, pipeline_details in pipelines.items():
                table_data.append(
                    [
                        type,
                        pipeline_details.get("name"),
                        pipeline_details.get("trigger"),
                        pipeline_details.get("pipeline_execution_id"),
                        pipeline_details.get("start_time"),
                        pipeline_details.get("status"),
                        pipeline_details.get("revision_id"),
                        pipeline_details.get("revision_summary"),
                        pipeline_details.get("duration"),
                        ", ".join(
                            [
                                f"{a.get('status')}"
                                for a in pipeline_details.get("trend", [])
                            ]
                        ),
                    ]
                )
        table = terminaltables.AsciiTable(table_data)
        click.echo(table.table)
    elif format == "html":
        content = ""
        content += "<table class='pipes'>"
        content += "    <tr>"
        content += "        <th>Name</th>"
        content += "        <th>Trigger</th>"
        content += "        <th>Status</th>"
        content += "        <th>Start Time</th>"
        content += "        <th>Revision Id</th>"
        content += "        <th>Revision Summary</th>"
        content += "        <th>Duration</th>"
        content += "        <th>Trend</th>"
        content += "    </tr>"

        for type, pipelines in pipelines_to_check.items():
            for pipeline_name, pipeline_details in pipelines.items():
                result = result_look_up.get(pipeline_details.get("status"), "amber")
                trend = "<br /> ".join(
                    [
                        f"{a.get('trigger')}:{a.get('status')[0]}"
                        for a in pipeline_details.get("trend", [])
                    ]
                )
                content += f"    <tr class='{result}'>"
                content += f"        <td>{pipeline_details.get('name')}</td>"
                content += f"        <td>{pipeline_details.get('status')}</td>"
                content += f"        <td>{pipeline_details.get('trigger')}</td>"
                content += f"        <td>{pipeline_details.get('start_time')}</td>"
                content += f"        <td>{pipeline_details.get('revision_id')}</td>"
                content += (
                    f"        <td>{pipeline_details.get('revision_summary')}</td>"
                )
                content += f"        <td>{pipeline_details.get('duration')}</td>"
                content += f"        <td>{trend}</td>"
                content += "    </tr>"

        content += "</table>"
        template = utils.ENV.get_template(constants.STATIC_HTML_PAGE)
        rendered = template.render(title="Factory Pipelines", content=content)
        click.echo(rendered)



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import json
import logging
import operator

import click
from betterboto import client as betterboto_client

from servicecatalog_puppet import constants


logger = logging.getLogger(constants.PUPPET_LOGGER_NAME)


def get_codebuild_build_ids_for(codebuild_project_name, limit):
    codebuild_ids = list()
    with betterboto_client.ClientContextManager("codebuild") as codebuild:
        paginator = codebuild.get_paginator("list_builds_for_project")
        iterator = paginator.paginate(
            projectName=codebuild_project_name,
            sortOrder="DESCENDING",
        )
        for page in iterator:
            for i in page["ids"]:
                codebuild_ids.append(i)
                if len(codebuild_ids) == limit:
                    return codebuild_ids
    return codebuild_ids


def get_codebuild_build_ids(filter, limit):
    codebuild_build_list = []
    logger.info(f"Getting all Build Ids for CodeBuild project")
    if filter == "none":
        codebuild_build_list += get_codebuild_build_ids_for(
            constants.FULL_RUN_CODEBUILD_PROJECT_NAME, limit
        )
        codebuild_build_list += get_codebuild_build_ids_for(
            constants.SINGLE_ACCOUNT_RUN_CODEBUILD_PROJECT_NAME, limit
        )
    elif filter == "full-runs":
        codebuild_build_list += get_codebuild_build_ids_for(
            constants.FULL_RUN_CODEBUILD_PROJECT_NAME, limit
        )
    elif filter == "single-runs":
        codebuild_build_list += get_codebuild_build_ids_for(
            constants.SINGLE_ACCOUNT_RUN_CODEBUILD_PROJECT_NAME, limit
        )
    return codebuild_build_list


def get_build_info(build_id_list: list):
    """
    Returns the CodeBuild build information for the supplied CodeBuild build IDds.

        Parameters:
            build_id_list (list): List of CodeBuild build Ids

        Returns:
            build_info (list): List of dict objects with the information for each of the supplied build Ids
    """
    build_info = []
    logger.info(f"Getting information for {len(build_id_list)} builds")
    while len(build_id_list) > 0:
        logger.info(f"list length {len(build_id_list)}")
        inner_build_list = build_id_list[:50]

        with betterboto_client.ClientContextManager("codebuild") as codebuild:
            response = codebuild.batch_get_builds(ids=inner_build_list)

            for build in response["builds"]:
                duration = build["endTime"] - build["startTime"]
                inner_response_dict = {
                    "id": build["id"],
                    "start": build["startTime"],
                    "duration_in_seconds": str(round(duration.total_seconds())),
                    "status": build["buildStatus"],
                }
                build_info.append(inner_response_dict)
            del build_id_list[:50]
    return build_info


def write_csv(build_details):
    """
    Writes a CSV file with the build information from the supplied list

        Parameters:
            build_details (list): List of dict objects with the information for each build Id
            file_name (str): Name of the CSV file to create
    """
    logger.info(f"Writing output to CSV")
    results = [",".join(["id", "startTime", "duration_in_seconds", "status"])]
    for build in build_details:
        results.append(
            ",".join(
                [
                    build["id"],
                    build["start"].strftime("%d/%m/%Y %H:%M:%S"),
                    build["duration_in_seconds"],
                    build["status"],
                ]
            )
        )
    return "\n".join(results)


def sort_and_trim(what, limit):
    result = sorted(what, key=operator.itemgetter("start"), reverse=True)
    return result[:limit]


def show_codebuilds(filter, limit, format):
    """
    Exports the CodeBuild job statistics to CSV for the single account Puppet execution
    """
    build_ids = get_codebuild_build_ids(filter, limit)
    build_details_single = get_build_info(build_ids)
    result = sort_and_trim(build_details_single, limit)
    if format == "csv":
        click.echo(write_csv(result))
    elif format == "json":
        click.echo(json.dumps(result, indent=4, default=str))



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from unittest import skip

from servicecatalog_puppet.workflow import tasks_unit_tests_helper


class DoExecuteTagPoliciesTaskTest(tasks_unit_tests_helper.PuppetTaskUnitTest):
    tag_policy_name = "tag_policy_name"
    region = "region"
    account_id = "account_id"
    ou_name = "ou_name"
    content = {}
    description = "description"
    manifest_file_path = "manifest_file_path"
    requested_priority = 1
    get_or_create_policy_ref = "get_or_create_policy_ref"

    def setUp(self) -> None:
        from servicecatalog_puppet.workflow.tag_policies import (
            do_execute_tag_policies_task,
        )

        self.module = do_execute_tag_policies_task

        self.sut = self.module.DoExecuteTagPoliciesTask(
            **self.get_common_args(),
            tag_policy_name=self.tag_policy_name,
            region=self.region,
            account_id=self.account_id,
            ou_name=self.ou_name,
            content=self.content,
            description=self.description,
            manifest_file_path=self.manifest_file_path,
            requested_priority=self.requested_priority,
            get_or_create_policy_ref=self.get_or_create_policy_ref,
        )

        self.wire_up_mocks()

    def test_params_for_results_display(self):
        # setup
        expected_result = {
            "puppet_account_id": self.puppet_account_id,
            "tag_policy_name": self.tag_policy_name,
            "region": self.region,
            "account_id": self.account_id,
            "ou_name": self.ou_name,
        }

        # exercise
        actual_result = self.sut.params_for_results_display()

        # verify
        self.assertEqual(expected_result, actual_result)

    @skip
    def test_requires(self):
        # setup
        # exercise
        actual_result = self.sut.requires()

        # verify
        raise NotImplementedError()

    @skip
    def test_run(self):
        # setup
        # exercise
        actual_result = self.sut.run()

        # verify
        raise NotImplementedError()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
from servicecatalog_puppet import constants, serialisation_utils
from servicecatalog_puppet.workflow import workflow_utils


SECTIONS_THAT_MUST_RUN_IN_HUB = [
    constants.PORTFOLIO_SHARE_AND_ACCEPT_ACCOUNT,
]
SECTIONS_THAT_MUST_RUN_IN_HUB_FOR_PUPPET_AS_A_SPOKE = [
    constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION,
    constants.PORTFOLIO_GET_ALL_PRODUCTS_AND_THEIR_VERSIONS,
]


def task_should_execute_before_spoke_execution(task, puppet_account_id):
    if task.get("section_name") in SECTIONS_THAT_MUST_RUN_IN_HUB:
        return True

    if (
        task.get("section_name") in SECTIONS_THAT_MUST_RUN_IN_HUB_FOR_PUPPET_AS_A_SPOKE
        and task.get("account_id") == puppet_account_id
    ):
        return True

    return False


def generate(puppet_account_id, all_tasks, output_file_path):
    tasks_to_include = dict()
    generate_manifest_ref = constants.GENERATE_MANIFEST
    for task_name, task in all_tasks.get("all_tasks").items():
        execution = task.get("execution", constants.EXECUTION_MODE_DEFAULT)
        if execution in [constants.EXECUTION_MODE_HUB, constants.EXECUTION_MODE_ASYNC]:
            should_include = True
        elif execution == constants.EXECUTION_MODE_SPOKE:
            # should_include = False
            # sharing should happen from the hub for launches in spoke mode
            should_include = task_should_execute_before_spoke_execution(
                task, puppet_account_id
            )

            # if task.get("section_name") == constants.PORTFOLIO_SHARE_AND_ACCEPT_ACCOUNT:
            #     should_include = True
            # else:
            #     if task.get("section_name") in SECTIONS_THAT_MUST_RUN_IN_HUB and task.get("account_id") == puppet_account_id:
            #         should_include = True
            #     else:
            #         should_include = False

        elif execution == constants.EXECUTION_MODE_HUB_AND_SPOKE_SPLIT:
            # cannot assume account_id role from spoke when it is the puppet account id
            should_include = (
                task.get("account_id", puppet_account_id) == puppet_account_id
            )
            # these should not override the previous decisions
            if not should_include:
                # sharing should happen from the hub for spoke-local-portfolios in hub and spoke split mode
                should_include = task.get("section_name") in [
                    constants.PORTFOLIO_SHARE_AND_ACCEPT_ACCOUNT,
                    constants.PORTFOLIO_SHARE_AND_ACCEPT_AWS_ORGANIZATIONS,
                ]
        else:
            raise Exception(f"Unhandled execution: {execution}")

        if should_include:
            tasks_to_include[task_name] = task
        else:
            if not tasks_to_include.get(generate_manifest_ref):
                tasks_to_include[generate_manifest_ref] = dict(
                    puppet_account_id=puppet_account_id,
                    task_reference=generate_manifest_ref,
                    section_name=constants.GENERATE_MANIFEST,
                    dependencies_by_reference=[],
                )

            replacement_ref = (
                f"{constants.RUN_DEPLOY_IN_SPOKE}_{task.get('account_id')}"
            )
            if not tasks_to_include.get(replacement_ref):
                tasks_to_include[replacement_ref] = dict(
                    execution=constants.EXECUTION_MODE_HUB,
                    puppet_account_id=puppet_account_id,
                    account_id=task.get("account_id"),
                    section_name=constants.RUN_DEPLOY_IN_SPOKE,
                    task_reference=replacement_ref,
                    generate_manifest_ref=generate_manifest_ref,
                    dependencies_by_reference=[
                        generate_manifest_ref,
                    ],
                )

    # make sure everything runs before we zip up the output directory
    if tasks_to_include.get(generate_manifest_ref):
        t = tasks_to_include[generate_manifest_ref]
        for task_name, task_to_include in tasks_to_include.items():
            if task_to_include.get("section_name") not in [
                constants.RUN_DEPLOY_IN_SPOKE,
                constants.GENERATE_MANIFEST,
            ]:
                t["dependencies_by_reference"].append(task_name)

    for task_name, task_to_include in tasks_to_include.items():
        for dep in task_to_include.get("dependencies_by_reference"):
            if not tasks_to_include.get(dep):
                raise Exception(
                    f"{task_name} depends on: {dep} which is not listed in this reference"
                )

    result = dict(all_tasks=tasks_to_include)
    workflow_utils.ensure_no_cyclic_dependencies("hub task reference", tasks_to_include)
    # open(output_file_path, "w").write(serialisation_utils.dump(result))
    open(output_file_path, "w").write(serialisation_utils.dump_as_json(result))
    return result



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
from servicecatalog_puppet import constants, task_reference_constants


def handle_tag_policies(
    all_tasks,
    all_tasks_task_reference,
    item_name,
    puppet_account_id,
    section_name,
    task_reference,
    task_to_add,
):
    get_or_create_policy_ref = (
        f"{constants.GET_OR_CREATE_TAG_POLICIES_POLICY}-{item_name}"
    )
    if not all_tasks.get(get_or_create_policy_ref):
        all_tasks[get_or_create_policy_ref] = {
            "task_reference": get_or_create_policy_ref,
            "execution": "hub",
            "account_id": puppet_account_id,
            "region": task_to_add.get("region"),
            "policy_name": task_to_add.get("tag_policy_name"),
            "policy_description": task_to_add.get("description"),
            "policy_content": task_to_add.get("content"),
            "dependencies_by_reference": list(),
            task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
            task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
            task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
            "section_name": constants.GET_OR_CREATE_TAG_POLICIES_POLICY,
        }
    all_tasks[all_tasks_task_reference][
        "get_or_create_policy_ref"
    ] = get_or_create_policy_ref
    all_tasks[all_tasks_task_reference]["dependencies_by_reference"].append(
        get_or_create_policy_ref
    )
    all_tasks[get_or_create_policy_ref][
        task_reference_constants.MANIFEST_SECTION_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
    all_tasks[get_or_create_policy_ref][
        task_reference_constants.MANIFEST_ITEM_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
    all_tasks[get_or_create_policy_ref][
        task_reference_constants.MANIFEST_ACCOUNT_IDS
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import functools

import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks


class DoExecuteServiceControlPoliciesTask(tasks.TaskWithReference):
    service_control_policy_name = luigi.Parameter()
    region = luigi.Parameter()
    account_id = luigi.Parameter()
    ou_name = luigi.Parameter()
    get_or_create_policy_ref = luigi.Parameter()

    content = luigi.DictParameter()
    description = luigi.Parameter()

    requested_priority = luigi.IntParameter()

    manifest_file_path = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "service_control_policy_name": self.service_control_policy_name,
            "region": self.region,
            "account_id": self.account_id,
            "ou_name": self.ou_name,
        }

    @functools.lru_cache(maxsize=32)
    def target(self):
        with self.organizations_policy_client() as orgs:
            if self.account_id != "":
                return self.account_id
            else:
                if str(self.ou_name).startswith("/"):
                    return orgs.convert_path_to_ou(self.ou_name)
                else:
                    return self.ou_name

    def has_policy_attached(self, orgs):
        paginator = orgs.get_paginator("list_policies_for_target")
        for page in paginator.paginate(
            TargetId=self.target(), Filter="SERVICE_CONTROL_POLICY"
        ):
            for policy in page.get("Policies", []):
                if policy.get("Name") == self.service_control_policy_name:
                    return True
        return False

    def run(self):
        with self.organizations_policy_client() as orgs:
            self.info("Ensuring attachments for policies")
            policy_id = self.get_output_from_reference_dependency(
                self.get_or_create_policy_ref
            ).get("Id")
            if self.has_policy_attached(orgs):
                self.write_empty_output()
            else:
                orgs.attach_policy(PolicyId=policy_id, TargetId=self.target())
                self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import json

import luigi

import servicecatalog_puppet.manifest_utils
import servicecatalog_puppet.serialisation_utils
from servicecatalog_puppet import constants, serialisation_utils
from servicecatalog_puppet.workflow.dependencies import tasks


class GetOrCreatePolicyTask(tasks.TaskWithReference):
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    policy_name = luigi.Parameter()
    policy_description = luigi.Parameter()
    policy_content = luigi.DictParameter()
    tags = luigi.ListParameter()

    manifest_file_path = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "account_id": self.account_id,
            "policy_name": self.policy_name,
        }

    def get_unwrapped_policy(self):
        if self.policy_content.get("default") is not None:
            unwrapped = servicecatalog_puppet.serialisation_utils.unwrap(
                self.policy_content.get("default")
            )
        elif self.policy_content.get("s3") is not None:
            with self.spoke_client("s3") as s3:
                bucket = self.policy_content.get("s3").get("bucket")
                key = self.policy_content.get("s3").get("key")
                raw_data = (
                    s3.get_object(Bucket=bucket, Key=key)
                    .get("Body")
                    .read()
                    .decode("utf-8")
                )
                unwrapped = serialisation_utils.json_loads(raw_data)
        else:
            raise Exception("Not supported policy content structure")
        return unwrapped

    def run(self):
        with self.organizations_policy_client() as orgs:
            unwrapped = self.get_unwrapped_policy()
            content = json.dumps(unwrapped, indent=0, default=str)

            tags = [dict(Key="ServiceCatalogPuppet:Actor", Value="generated")]
            for tag in self.tags:
                tags.append(dict(Key=tag.get("Key"), Value=tag.get("Value")))

            paginator = orgs.get_paginator("list_policies")
            for page in paginator.paginate(Filter="SERVICE_CONTROL_POLICY"):
                for policy in page.get("Policies", []):
                    if policy.get("Name") == self.policy_name:
                        kwargs = dict(PolicyId=policy.get("Id"))

                        if policy.get("Description") != self.policy_description:
                            kwargs["Description"] = self.policy_description

                        remote_policy_content = (
                            orgs.describe_policy(PolicyId=policy.get("Id"))
                            .get("Policy")
                            .get("Content")
                        )

                        if unwrapped != serialisation_utils.json_loads(
                            remote_policy_content
                        ):
                            kwargs["Content"] = content

                        if len(kwargs.keys()) > 1:
                            result = (
                                orgs.update_policy(**kwargs)
                                .get("Policy")
                                .get("PolicySummary")
                            )
                            self.write_output(result)
                            return
                        else:
                            self.write_output(policy)
                            return

            result = (
                orgs.create_policy(
                    Name=self.policy_name,
                    Description=self.policy_description,
                    Type="SERVICE_CONTROL_POLICY",
                    Tags=tags,
                    Content=content,
                )
                .get("Policy")
                .get("PolicySummary")
            )
            self.write_output(result)



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import json

import orjson
import yaml

from servicecatalog_puppet import serialisation_utils


class Equals(yaml.YAMLObject):
    yaml_loader = yaml.SafeLoader
    yaml_dumper = yaml.SafeDumper

    yaml_tag = "!Equals"

    def __init__(self, values):
        self.values = values

    @classmethod
    def from_yaml(cls, constructor, node):
        x = constructor.construct_sequence(node)
        return cls(values=x)

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_sequence(cls.yaml_tag, data.values)

    def get_result(self):
        return all(element == self.values[0] for element in self.values)


class Not(yaml.YAMLObject):
    yaml_loader = yaml.SafeLoader
    yaml_dumper = yaml.SafeDumper

    yaml_tag = "!Not"

    def __init__(self, values):
        self.values = values

    @classmethod
    def from_yaml(cls, constructor, node):
        x = constructor.construct_sequence(node)
        return cls(values=x)

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_sequence(cls.yaml_tag, data.values)

    def get_result(self):
        return not self.values[0]


def load(what):
    return yaml.safe_load(what)


def dump(what):
    return yaml.safe_dump(what, default_flow_style=False, width=1000000)


def dump_as_json(input):
    return json.dumps(input, default=str)


def load_as_json(input):
    return serialisation_utils.json_loads(input)


def json_dumps(obj):
    return orjson.dumps(obj, option=orjson.OPT_INDENT_2)


def json_loads(s):
    return orjson.loads(s)


def unwrap(what):
    if hasattr(what, "get_wrapped"):
        return unwrap(what.get_wrapped())

    if isinstance(what, dict):
        thing = dict()
        for k, v in what.items():
            thing[k] = unwrap(v)
        return thing

    if isinstance(what, tuple):
        thing = list()
        for v in what:
            thing.append(unwrap(v))
        return thing

    if isinstance(what, list):
        thing = list()
        for v in what:
            thing.append(unwrap(v))
        return thing

    return what



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import io
import zipfile

import luigi

from servicecatalog_puppet import config, constants, serialisation_utils
from servicecatalog_puppet.serialisation_utils import unwrap
from servicecatalog_puppet.workflow.dependencies import tasks


class DeployC7NPolicies(tasks.TaskWithReferenceAndCommonParameters):
    role_name = luigi.Parameter()
    role_path = luigi.Parameter()
    policies = luigi.ListParameter()
    deployments = luigi.DictParameter()
    uses_orgs = luigi.BoolParameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
        }

    def generate_policies(self, partition):
        policies = list()
        member_role = (
            "arn:"
            + partition
            + ":iam::{account_id}:role"
            + self.role_path
            + self.role_name
        )
        for policy in unwrap(self.policies):
            if policy.get("mode", {}).get("type") == "cloudtrail":
                policy["mode"]["member-role"] = member_role
            policies.append(policy)
        return serialisation_utils.dump(unwrap(dict(policies=policies)))

    def generate_accounts(self, partition):
        accounts = dict()
        for deployment_region, deployment_accounts in self.deployments.items():
            for deployment_account_id in deployment_accounts:
                if accounts.get(deployment_account_id) is None:
                    accounts[deployment_account_id] = []
                accounts[deployment_account_id].append(deployment_region)

        result = []
        for account_id, regions in accounts.items():
            result.append(
                {
                    "account_id": account_id,
                    "name": str(account_id),
                    "regions": regions,
                    "role": f"arn:{partition}:iam::{account_id}:role{self.role_path}{self.role_name}",
                }
            )

        return serialisation_utils.dump(unwrap(dict(accounts=result)))

    def run(self):
        partition = config.get_partition()
        policies = self.generate_policies(partition)
        accounts = self.generate_accounts(partition)

        zip_buffer = io.BytesIO()
        with zipfile.ZipFile(zip_buffer, "a", zipfile.ZIP_DEFLATED, False) as zip_file:
            zip_file.writestr("accounts.yaml", accounts)
            zip_file.writestr("policies.yaml", policies)

            # ('2.txt', io.BytesIO(b'222'))]:

        bucket = f"sc-puppet-c7n-artifacts-{self.account_id}-{self.region}"
        key = "latest"

        with self.spoke_regional_client("s3") as s3:
            s3.put_object(
                Bucket=bucket,
                Key=key,
                Body=zip_buffer.getvalue(),
            )
        custodian_role_arn = (
            f"arn:{partition}:iam::{self.account_id}:role"
            + self.role_path
            + self.role_name
        )
        regions_to_run_in = list(self.deployments.keys())

        with self.spoke_client("ssm") as ssm:
            ssm.put_parameter(
                Name="/servicecatalog-puppet/aws-c7n-lambdas/REGIONS",
                Value=" ".join(regions_to_run_in),
                Type="String",
                Overwrite=True,
            )
            ssm.put_parameter(
                Name="/servicecatalog-puppet/aws-c7n-lambdas/CUSTODIAN_ROLE_ARN",
                Value=custodian_role_arn,
                Type="String",
                Overwrite=True,
            )
        with self.spoke_client("codebuild") as codebuild:
            result = codebuild.start_build_and_wait_for_completion(
                projectName="servicecatalog-puppet-deploy-c7n",
            )
            if result.get("buildStatus") != "SUCCEEDED":
                raise Exception(f"Deploying policy failed: {result.get('buildStatus')}")

        self.write_empty_output()



#   Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#   SPDX-License-Identifier: Apache-2.0
from unittest import mock


@mock.patch("click.secho")
def test_echo(mocked_click_secho):
    # setup
    from servicecatalog_puppet import print_utils

    message = "hello world"

    # exercise
    print_utils.echo(message)

    # verify
    mocked_click_secho.assert_called_once_with(message)


@mock.patch("click.secho")
def test_warn(mocked_click_secho):
    from servicecatalog_puppet import print_utils

    message = "hello world"

    # exercise
    print_utils.warn(message)

    # verify
    mocked_click_secho.assert_called_once_with(
        message,
        err=True,
        fg="yellow",
    )


@mock.patch("click.secho")
def test_error(mocked_click_secho):
    from servicecatalog_puppet import print_utils

    message = "hello world"

    # exercise
    print_utils.error(message)

    # verify
    mocked_click_secho.assert_called_once_with(
        message,
        err=True,
        fg="red",
    )



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import glob
import os
import shutil
import zipfile
from datetime import datetime, timedelta
from pathlib import Path

import cfn_tools
import click
import terminaltables
import yaml
from betterboto import client as betterboto_client
from jinja2 import Template

from servicecatalog_puppet import (
    asset_helpers,
    config,
    constants,
    print_utils,
    serialisation_utils,
    viz_template,
)


def upload_config(config):
    with betterboto_client.ClientContextManager("ssm") as ssm:
        ssm.put_parameter(
            Name=constants.CONFIG_PARAM_NAME,
            Type="String",
            Value=yaml.safe_dump(config),
            Overwrite=True,
        )
    click.echo("Uploaded config")


def seed(complexity, p):
    example = "manifest-{}.yaml".format(complexity)
    shutil.copy2(
        asset_helpers.resolve_from_site_packages(
            os.path.sep.join(["manifests", example])
        ),
        os.path.sep.join([p, "manifest.yaml"]),
    )


def list_resources():
    click.echo("# Framework resources")

    click.echo("## SSM Parameters used")
    click.echo(f"- {constants.CONFIG_PARAM_NAME}")
    click.echo(f"- {constants.CONFIG_PARAM_NAME_ORG_IAM_ROLE_ARN}")

    for file in Path(__file__).parent.resolve().glob("*.template.yaml"):
        if "empty.template.yaml" == file.name:
            continue
        template_contents = Template(open(file, "r").read()).render()
        template = cfn_tools.load_yaml(template_contents)
        click.echo(f"## Resources for stack: {file.name.split('.')[0]}")
        table_data = [
            [
                "Logical Name",
                "Resource Type",
                "Name",
            ],
        ]
        table = terminaltables.AsciiTable(table_data)
        for logical_name, resource in template.get("Resources").items():
            resource_type = resource.get("Type")
            name = "-"
            type_to_name = {
                "AWS::IAM::Role": "RoleName",
                "AWS::SSM::Parameter": "Name",
                "AWS::S3::Bucket": "BucketName",
                "AWS::CodePipeline::Pipeline": "Name",
                "AWS::CodeBuild::Project": "Name",
                "AWS::CodeCommit::Repository": "RepositoryName",
                "AWS::SNS::Topic": "TopicName",
                "AWS::SQS::Queue": "QueueName",
            }

            if type_to_name.get(resource_type) is not None:
                name = resource.get("Properties", {}).get(
                    type_to_name.get(resource_type), "Not Specified"
                )
                if not isinstance(name, str):
                    name = cfn_tools.dump_yaml(name)

            table_data.append([logical_name, resource_type, name])

        click.echo(table.table)
    click.echo(f"n.b. AWS::StackName evaluates to {constants.BOOTSTRAP_STACK_NAME}")


def set_config_value(name, value, is_a_boolean=True):
    with betterboto_client.ClientContextManager(
        "ssm", region_name=constants.HOME_REGION
    ) as ssm:
        try:
            response = ssm.get_parameter(Name=constants.CONFIG_PARAM_NAME)
            config = yaml.safe_load(response.get("Parameter").get("Value"))
        except ssm.exceptions.ParameterNotFound:
            config = {}

        if name == "regions":
            config["regions"] = value if len(value) > 1 else value[0].split(",")
        else:
            if is_a_boolean:
                config[name] = value.upper() == "TRUE"
            else:
                config[name] = value

        upload_config(config)


def set_named_config_value(name, value):
    with betterboto_client.ClientContextManager(
        "ssm", region_name=constants.HOME_REGION
    ) as ssm:
        ssm.put_parameter(
            Name=name,
            Type="String",
            Value=value,
            Overwrite=True,
        )
        click.echo("Uploaded named config")


def handle_action_execution_detail(puppet_account_id, action_execution_detail):
    action_type_id = action_execution_detail.get("input").get("actionTypeId")
    if (
        action_type_id.get("category") == "Build"
        and action_type_id.get("owner") == "AWS"
        and action_type_id.get("provider") == "CodeBuild"
    ):
        external_execution_id = (
            action_execution_detail.get("output")
            .get("executionResult")
            .get("externalExecutionId")
        )

        with betterboto_client.ClientContextManager(
            "codebuild", region_name=config.get_home_region(puppet_account_id)
        ) as codebuild:
            builds = codebuild.batch_get_builds(ids=[external_execution_id]).get(
                "builds"
            )
            build = builds[0]
            log_details = build.get("logs")
            with betterboto_client.ClientContextManager(
                "logs", region_name=config.get_home_region(puppet_account_id)
            ) as logs:
                project_name = (
                    action_execution_detail.get("input")
                    .get("configuration")
                    .get("ProjectName")
                )
                action_execution_id = action_execution_detail.get("actionExecutionId")
                output_file_name = f"log-{project_name}--{action_execution_id}.log"
                with open(
                    output_file_name,
                    "w",
                ) as f:
                    params = {
                        "logGroupName": log_details.get("groupName"),
                        "logStreamName": log_details.get("streamName"),
                        "startFromHead": True,
                    }
                    has_more_logs = True
                    while has_more_logs:
                        get_log_events_response = logs.get_log_events(**params)
                        if (len(get_log_events_response.get("events"))) > 0:
                            params["nextToken"] = get_log_events_response.get(
                                "nextForwardToken"
                            )
                        else:
                            has_more_logs = False
                            if params.get("nextToken"):
                                del params["nextToken"]
                        for e in get_log_events_response.get("events"):
                            d = datetime.utcfromtimestamp(
                                e.get("timestamp") / 1000
                            ).strftime("%Y-%m-%d %H:%M:%S")
                            f.write(f"{d} : {e.get('message')}")


def export_puppet_pipeline_logs(execution_id, puppet_account_id):
    with betterboto_client.ClientContextManager(
        "codepipeline", region_name=config.get_home_region(puppet_account_id)
    ) as codepipeline:
        action_execution_details = codepipeline.list_action_executions(
            pipelineName=constants.PIPELINE_NAME,
            filter={"pipelineExecutionId": execution_id},
        ).get("actionExecutionDetails")

        for action_execution_detail in action_execution_details:
            handle_action_execution_detail(puppet_account_id, action_execution_detail)


def get_data_for_viz(codebuild_build_id, output_file_name_prefix):
    with betterboto_client.ClientContextManager("codebuild") as codebuild:
        build = codebuild.batch_get_builds(ids=[codebuild_build_id]).get("builds")[0]
        location = build.get("artifacts").get("location").split(":")
        bucket = location[5].split("/")[0]
        key = location[5].replace(f"{bucket}/", "")
        zip_file_location = f"{output_file_name_prefix}.zip"
        if os.path.exists(zip_file_location):
            print(f"Found zip file, skipping download")
        else:
            print(f"Downloading zip file")
            with betterboto_client.ClientContextManager("s3") as s3:
                print(f"getting {bucket} {key}")
                s3.download_file(Bucket=bucket, Key=key, Filename=zip_file_location)
        if os.path.exists(output_file_name_prefix):
            print(f"Found output folder, skipping unzip")
        else:
            print(f"Unziping")
            os.makedirs(output_file_name_prefix)
            with zipfile.ZipFile(zip_file_location, "r") as zip_ref:
                zip_ref.extractall(output_file_name_prefix)
    return f"{output_file_name_prefix}/results"


def generate_data_for_viz(path_to_results, group_by_pid):
    results = list()
    groups = dict()
    time_format = "%Y-%m-%d %H:%M:%S"
    earliest_time = datetime.strptime("4022-09-14 00:54:33", time_format)
    latest_time = datetime.strptime("2000-09-14 00:54:33", time_format)
    for starter in glob.glob(f"{path_to_results}/start/*.json"):
        start = serialisation_utils.json_loads(open(starter, "r").read())
        name = os.path.basename(starter)
        end = None
        task_id = name.replace(".json", "")
        task_id = "-".join(task_id.split("-")[1:])
        processing_time = dict(duration="unknown")

        if os.path.exists(f"{path_to_results}/success/{name}"):
            result = "success"
            end = serialisation_utils.json_loads(
                open(f"{path_to_results}/success/{name}", "r").read()
            )
        elif os.path.exists(f"{path_to_results}/failure/{name}"):
            result = "failure"
            end = serialisation_utils.json_loads(
                open(f"{path_to_results}/failure/{name}", "r").read()
            )

        if os.path.exists(f"{path_to_results}/processing_time/{name}"):
            processing_time = serialisation_utils.json_loads(
                open(f"{path_to_results}/processing_time/{name}", "r").read()
            )

        if end:
            body = ""
            starting_time = start.get("datetime")
            ending_time = end.get("datetime")
            pid = start.get("pid")
            groups[pid] = dict(
                id=pid,
                content=pid,
                value=pid,
            )
            for name2, value2 in start.get("params_for_results", {}).items():
                body += f"<b>{name2}</b>:{value2}<br />"
            body += f"<b>pid</b>:{pid}<br />"
            body += f"<b>start</b>:{starting_time}<br />"
            body += f"<b>end</b>:{ending_time}<br />"
            body += f"<b>duration</b>:{processing_time.get('duration')}<br />"
            content = f"""<h2>{task_id}</h2><dl>{body}</dl>"""
            task_type = task_id.split("_")[0]
            class_name = f"{result} {task_type}"

            starting_time = datetime.strptime(starting_time, time_format)
            ending_time = datetime.strptime(ending_time, time_format)

            if starting_time < earliest_time:
                earliest_time = starting_time

            if ending_time > latest_time:
                latest_time = ending_time

            if group_by_pid:
                d = dict(
                    id=task_id,
                    group=pid,
                    content=task_id,
                    title=content,
                    start=start.get("datetime"),
                    end=end.get("datetime"),
                    className=class_name,
                )
            else:
                d = dict(
                    id=task_id,
                    content=task_id,
                    title=content,
                    start=start.get("datetime"),
                    end=end.get("datetime"),
                    className=class_name,
                )
            results.append(d)
        else:
            print_utils.warn(f"Did not find an end for {name}")

    earliest_time = earliest_time - timedelta(minutes=2)
    latest_time = latest_time + timedelta(minutes=2)
    groups = list(groups.values())
    return (
        results,
        groups,
        earliest_time.strftime(time_format),
        latest_time.strftime(time_format),
    )


def export_deploy_viz(codebuild_execution_id, group_by_pid, puppet_account_id):
    output_file_name_prefix = codebuild_execution_id.split(":")[1]
    path_to_results = get_data_for_viz(codebuild_execution_id, output_file_name_prefix)
    results, groups, start, end = generate_data_for_viz(path_to_results, group_by_pid)
    if group_by_pid:
        params = "container, items, groups, options"
    else:
        params = "container, items, options"
    output = viz_template.CONTENT.format(
        DATASET=results,
        START=start,
        END=end,
        GROUPS=groups,
        PARAMS=params,
    )
    f = open(f"{output_file_name_prefix}.html", "w")
    f.write(output)
    f.close()


def generate_data_for_traces(path_to_results):
    results = list()
    for f in glob.glob(f"{path_to_results}/traces/*.json"):
        results.append(open(f, "r").read())
    return f'{{"traceEvents": [{",".join(results)}]}}'


def export_traces(codebuild_execution_id, puppet_account_id):
    bucket = f"sc-puppet-log-store-{puppet_account_id}"
    key_prefix = f"{codebuild_execution_id}/traces"

    results = list()
    with betterboto_client.ClientContextManager("s3") as s3:
        paginator = s3.get_paginator("list_objects_v2")
        for page in paginator.paginate(
            Bucket=bucket,
            Prefix=key_prefix,
        ):
            for obj in page.get("Contents", []):
                key = obj.get("Key")
                results.append(
                    s3.get_object(Bucket=bucket, Key=key)
                    .get("Body")
                    .read()
                    .decode("utf-8")
                )

    traces = f'{{"traceEvents": [{",".join(results)}]}}'
    with open(f"{codebuild_execution_id.split('/')[-1]}-traces.json", "w") as f:
        f.write(traces)



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks


class ShareAndAcceptPortfolioForAccountTask(tasks.TaskWithReference):
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    portfolio = luigi.Parameter()
    share_tag_options = luigi.BoolParameter()
    share_principals = luigi.BoolParameter()
    portfolio_task_reference = luigi.Parameter()
    describe_portfolio_shares_task_ref = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "region": self.region,
            "account_id": self.account_id,
            "share_tag_options": self.share_tag_options,
            "share_principals": self.share_principals,
        }

    def accept_if_needed(self, portfolio_id):
        accepted = False
        with self.spoke_regional_client("servicecatalog") as servicecatalog:
            paginator = servicecatalog.get_paginator("list_accepted_portfolio_shares")
            for page in paginator.paginate(PortfolioShareType="IMPORTED"):
                for portfolio_detail in page.get("PortfolioDetails", []):
                    if portfolio_detail.get("Id") == portfolio_id:
                        accepted = True
        if not accepted:
            servicecatalog.accept_portfolio_share(PortfolioId=portfolio_id)
            return True
        return False

    def run(self):
        portfolio_id = self.get_attribute_from_output_from_reference_dependency(
            "Id", self.portfolio_task_reference
        )

        changes = dict()
        existing_share_details = (
            self.get_attribute_from_output_from_reference_dependency(
                self.account_id, self.describe_portfolio_shares_task_ref
            )
        )

        if existing_share_details:
            if existing_share_details.get("ShareTagOptions") != self.share_tag_options:
                changes["ShareTagOptions"] = self.share_tag_options

            if changes:
                with self.hub_regional_client("servicecatalog") as servicecatalog:
                    servicecatalog.update_portfolio_share(
                        PortfolioId=portfolio_id, AccountId=self.account_id, **changes
                    )
        else:
            # SHARE
            self.info(f"sharing {portfolio_id} with {self.account_id}")
            with self.hub_regional_client("servicecatalog") as servicecatalog:
                servicecatalog.create_portfolio_share(
                    PortfolioId=portfolio_id,
                    AccountId=self.account_id,
                    ShareTagOptions=self.share_tag_options,
                )

            # ACCEPT
            self.accept_if_needed(portfolio_id)

        self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks
from servicecatalog_puppet.workflow.portfolio.associations import association_utils


class TerminateAssociationsForSpokeLocalPortfolioTask(tasks.TaskWithReference):
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    portfolio = luigi.Parameter()
    spoke_local_portfolio_name = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "spoke_local_portfolio_name": self.spoke_local_portfolio_name,
            "region": self.region,
            "account_id": self.account_id,
        }

    def run(self):
        stack_name = (
            association_utils.generate_stack_name_for_associations_by_item_name(
                self.spoke_local_portfolio_name
            )
        )

        with self.spoke_regional_client("cloudformation") as cloudformation:
            self.info(f"About to delete the stack: {stack_name}")
            cloudformation.ensure_deleted(StackName=stack_name)
        self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import functools
import logging

import jmespath
import luigi
from deepmerge import always_merger

from servicecatalog_puppet import constants, manifest_utils, serialisation_utils
from servicecatalog_puppet.commands import graph
from servicecatalog_puppet.waluigi.task_mixins import (
    io_mixin,
    task_executor_mixin,
)
from servicecatalog_puppet.workflow.dependencies import task_factory
from servicecatalog_puppet.workflow.task_mixins import (
    client_mixin,
    env_var_mixin,
)


logger = logging.getLogger(constants.PUPPET_LOGGER_NAME)


class TaskWithReference(
    task_executor_mixin.TaskExecutorMixin,
    env_var_mixin.EnvVarMixin,
    io_mixin.IOMixin,
    luigi.Task,
    client_mixin.ClientMixin,
):
    task_reference = luigi.Parameter()
    manifest_task_reference_file_path = luigi.Parameter()
    dependencies_by_reference = luigi.ListParameter()
    puppet_account_id = luigi.Parameter()
    manifest_files_path = luigi.Parameter()

    cachable_level = constants.CACHE_LEVEL_DEFAULT

    def get_expanded_manifest_file_path(self):
        return f"{self.manifest_files_path}/manifest-expanded.yaml"

    def get_from_manifest(self, section_name, item_name):
        with open(self.get_expanded_manifest_file_path(), "r") as f:
            m = serialisation_utils.load(f.read())
            return m[section_name][item_name]

    def requires(self):
        return dict(reference_dependencies=self.dependencies_for_task_reference())

    def read_from_input(self, input_name):
        with self.input().get(input_name).open("rb") as f:
            return f.read()

    def load_from_input(self, input_name):
        return serialisation_utils.json_loads(self.read_from_input(input_name))

    def get_output_from_reference_dependency(self, reference):
        with self.input().get("reference_dependencies").get(reference).open("r") as f:
            content = f.read()
        return serialisation_utils.json_loads(content)

    def get_attribute_from_output_from_reference_dependency(self, attribute, reference):
        return self.get_output_from_reference_dependency(reference).get(attribute)

    def get_output_from_reference_dependency_raw(self, reference):
        f = self.input().get("reference_dependencies").get(reference).open("r")
        content = f.read()
        f.close()
        return content

    @functools.lru_cache(maxsize=32)
    def get_task_from_reference(self, task_reference):
        f = open(
            f"{self.manifest_files_path}/tasks/{graph.escape(task_reference)}.json", "r"
        )
        c = f.read()
        f.close()
        return serialisation_utils.load_as_json(c)

    @functools.lru_cache(maxsize=32)
    def dependencies_for_task_reference(self):
        dependencies = dict()

        this_task = self.get_task_from_reference(self.task_reference)
        if this_task is None:
            raise Exception(f"Did not find {self.task_reference} within reference")
        for dependency_by_reference in this_task.get("dependencies_by_reference", []):
            dependency_by_reference_params = self.get_task_from_reference(
                dependency_by_reference
            )
            if dependency_by_reference_params is None:
                raise Exception(
                    f"{self.task_reference} has a dependency: {dependency_by_reference} unsatisfied by the manifest task reference"
                )
            t_reference = dependency_by_reference_params.get("task_reference")
            dependencies[t_reference] = task_factory.create(
                self.manifest_files_path,
                self.manifest_task_reference_file_path,
                self.puppet_account_id,
                dependency_by_reference_params,
            )
        return dependencies

    def info(self, message):
        logger.info(f"{self.task_reference}: {message}")

    def debug(self, message):
        logger.debug(f"{self.task_reference}: {message}")

    def error(self, message):
        logger.error(f"{self.task_reference}: {message}")

    def warning(self, message):
        logger.warning(f"{self.task_reference}: {message}")


class TaskWithReferenceAndCommonParameters(TaskWithReference):
    region = luigi.Parameter()
    account_id = luigi.Parameter()


class TaskWithParameters(TaskWithReference):
    def get_merged_launch_account_and_manifest_parameters(self):
        content = open(self.manifest_file_path, "r").read()
        manifest = manifest_utils.Manifest(serialisation_utils.load(content))

        result = dict()
        launch_parameters = (
            manifest.get(self.section_name).get(self.item_name).get("parameters", {})
        )
        manifest_parameters = manifest.get("parameters")
        account_parameters = manifest.get_account(self.account_id).get("parameters")

        always_merger.merge(result, manifest_parameters)
        always_merger.merge(result, launch_parameters)
        always_merger.merge(result, account_parameters)

        for p_name, p_details in result.items():
            if p_details.get("ssm"):
                for v_name, v_details in p_details.get("ssm").items():
                    result[p_name]["ssm"][v_name] = (
                        str(result[p_name]["ssm"][v_name])
                        .replace("${AWS::AccountId}", self.account_id)
                        .replace("${AWS::Region}", self.region)
                    )

        return result

    def get_parameter_values(self):
        all_params = {}
        p = self.get_merged_launch_account_and_manifest_parameters()
        for param_name, param_details in p.items():
            if param_details.get("ssm"):
                requested_param_details = param_details.get("ssm")
                requested_param_region = requested_param_details.get(
                    "region", constants.HOME_REGION
                )
                requested_param_account_id = requested_param_details.get(
                    "account_id", self.puppet_account_id
                )
                requested_param_name = (
                    str(requested_param_details.get("name"))
                    .replace("${AWS::AccountId}", self.account_id)
                    .replace("${AWS::Region}", self.region)
                )

                if requested_param_details.get("path"):
                    required_task_reference = f"{constants.SSM_PARAMETERS_WITH_A_PATH}-{requested_param_account_id}-{requested_param_region}-{requested_param_details.get('path')}"
                else:
                    required_task_reference = f"{constants.SSM_PARAMETERS}-{requested_param_account_id}-{requested_param_region}-{requested_param_name}"

                parameter_task_output = self.get_output_from_reference_dependency(
                    required_task_reference
                )

                if parameter_task_output.get(requested_param_name):
                    all_params[param_name] = parameter_task_output.get(
                        requested_param_name,
                    ).get("Value")
                elif requested_param_details.get("default"):
                    all_params[param_name] = requested_param_details.get("default")
                else:
                    raise Exception(
                        f"Could not find parameter value and no default was set for {requested_param_name}"
                    )

            if param_details.get("boto3"):
                requested_param_details = param_details.get("boto3")
                boto3_task_account_id = requested_param_details.get("account_id")
                boto3_task_region = requested_param_details.get("region")
                if param_details.get("cloudformation_stack_output"):
                    task_ref = (
                        f"{constants.BOTO3_PARAMETERS}"
                        f"-cloudformation_stack_output"
                        f"-{param_details.get('cloudformation_stack_output').get('stack_name')}"
                        f"-{param_details.get('cloudformation_stack_output').get('output_key')}"
                        f"-{boto3_task_account_id}"
                        f"-{boto3_task_region}"
                    )

                elif param_details.get("servicecatalog_provisioned_product_output"):
                    task_ref = (
                        f"{constants.BOTO3_PARAMETERS}"
                        f"-servicecatalog_provisioned_product_output"
                        f"-{param_details.get('servicecatalog_provisioned_product_output').get('provisioned_product_name')}"
                        f"-{param_details.get('servicecatalog_provisioned_product_output').get('output_key')}"
                        f"-{boto3_task_account_id}"
                        f"-{boto3_task_region}"
                    )
                else:
                    task_ref = (
                        f"{constants.BOTO3_PARAMETERS}"
                        f"-{self.section_name}"
                        f"-{self.item_name}"
                        f"-{param_name}"
                        f"-{account_id_to_use_for_boto3_call}"
                        f"-{region_to_use_for_boto3_call}"
                    )

                task_ref = (
                    str(task_ref.replace("${AWS::AccountId}", self.account_id))
                    .replace("${AWS::PuppetAccountId}", self.puppet_account_id)
                    .replace("${AWS::Region}", self.region)
                )
                print(
                    f"Getting output for task: {task_ref} within task {self.task_reference}"
                )
                parameter_task_output = self.get_output_from_reference_dependency(
                    task_ref
                )
                all_params[param_name] = parameter_task_output

            if param_details.get("s3"):
                requested_param_details = param_details.get("s3")
                key = str(requested_param_details.get("key"))
                key = (
                    key.replace("${AWS::Region}", self.region)
                    .replace("${AWS::AccountId}", self.account_id)
                    .replace("${AWS::PuppetAccountId}", self.puppet_account_id)
                )

                task_ref = f"{constants.S3_GET_OBJECT}" f"-{key}"

                print(
                    f"Getting output for task: {task_ref} within task {self.task_reference}"
                )
                parameter_task_output = self.get_output_from_reference_dependency(
                    task_ref
                )
                jmespath_location = str(requested_param_details.get("jmespath"))
                jmespath_location = (
                    jmespath_location.replace("${AWS::Region}", self.region)
                    .replace("${AWS::AccountId}", self.account_id)
                    .replace("${AWS::PuppetAccountId}", self.puppet_account_id)
                )
                default = str(requested_param_details.get("default"))
                default = (
                    default.replace("${AWS::Region}", self.region)
                    .replace("${AWS::AccountId}", self.account_id)
                    .replace("${AWS::PuppetAccountId}", self.puppet_account_id)
                )

                parameter_task_output = jmespath.search(
                    jmespath_location, parameter_task_output
                )
                if parameter_task_output is None:
                    if default is None:
                        raise Exception(
                            "Could not find value in the s3 JSON object and there is no default value."
                            "Check your JMESPath is correct."
                        )
                    else:
                        parameter_task_output = default

                all_params[param_name] = parameter_task_output

            if param_details.get("default"):
                all_params[param_name] = (
                    str(param_details.get("default"))
                    .replace("${AWS::AccountId}", self.account_id)
                    .replace("${AWS::Region}", self.region)
                )
            if param_details.get("mapping"):
                content = open(self.manifest_file_path, "r").read()
                manifest = manifest_utils.Manifest(serialisation_utils.load(content))

                all_params[param_name] = manifest.get_mapping(
                    param_details.get("mapping"), self.account_id, self.region
                )
        return all_params



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import logging
import os

from servicecatalog_puppet import (
    config,
    constants,
    manifest_utils,
    serialisation_utils,
    task_reference_constants,
)
from servicecatalog_puppet.commands import graph
from servicecatalog_puppet.commands.task_reference_helpers import (
    complete_generator,
    hub_generator,
)
from servicecatalog_puppet.workflow import runner


logger = logging.getLogger(constants.PUPPET_LOGGER_NAME)


def generate_task_reference(f):
    path = os.path.dirname(f.name)
    puppet_account_id = config.get_puppet_account_id()

    content = open(f.name, "r").read()
    manifest = manifest_utils.Manifest(serialisation_utils.load(content))
    complete = complete_generator.generate(  # hub and spokes
        puppet_account_id,
        manifest,
        os.path.dirname(f.name),
        # f.name.replace("-expanded.yaml", "-task-reference-full.json"),
    )
    hub_tasks = hub_generator.generate(  # hub only
        puppet_account_id,
        complete,
        f.name.replace("-expanded.yaml", "-task-reference.json"),
    )
    task_output_path = f"{path}/tasks"
    if not os.path.exists(task_output_path):
        os.makedirs(task_output_path)

    for t_name, task in complete.get("all_tasks", {}).items():
        task_output_file_path = f"{task_output_path}/{graph.escape(t_name)}.json"
        task_output_content = serialisation_utils.dump_as_json(task)
        open(task_output_file_path, "w").write(task_output_content)

    for t_name, task in hub_tasks.get("all_tasks", {}).items():
        task_output_file_path = f"{task_output_path}/{graph.escape(t_name)}.json"
        task_output_content = serialisation_utils.dump_as_json(task)
        open(task_output_file_path, "w").write(task_output_content)


def deploy_from_task_reference(path):
    f = f"{path}/manifest-task-reference.json"
    tasks_to_run = list()
    tasks_to_run_filtered = dict()
    reference = serialisation_utils.load_as_json(open(f, "r").read())
    all_tasks = reference.get("all_tasks")

    num_workers = config.get_num_workers()
    puppet_account_id = config.get_puppet_account_id()
    single_account_id = config.get_single_account_id()

    for task_reference, task in all_tasks.items():
        if single_account_id:  # running single account mode
            print("Running in single account mode")
            task_section_name = task.get("section_name")
            task_account_id = task.get("account_id")
            spoke_execution = str(config.get_executor_account_id()) != str(
                puppet_account_id
            )
            if spoke_execution:  # is a spoke execution
                if (
                    task_account_id == single_account_id
                    and task_section_name != constants.RUN_DEPLOY_IN_SPOKE
                ):
                    tasks_to_run_filtered[task_reference] = task

            else:  # is running in a hub
                # need to continue any tasks that should not be executed in single account mode
                if task.get("task_reference") == constants.CREATE_POLICIES:
                    continue

                if task.get("section_name") == constants.SSM_OUTPUTS:
                    if not any(
                        [
                            task.get(task_reference_constants.MANIFEST_ACCOUNT_IDS).get(
                                str(single_account_id)
                            ),
                            task.get(task_reference_constants.MANIFEST_ACCOUNT_IDS).get(
                                str(puppet_account_id)
                            ),
                        ]
                    ):
                        continue

                    if str(task.get("task_generating_output_account_id")) not in [
                        str(puppet_account_id),
                        str(single_account_id),
                    ]:
                        continue

                else:
                    if task_account_id and str(task_account_id) not in [
                        str(single_account_id),
                        str(puppet_account_id),
                    ]:
                        continue

                tasks_to_run_filtered[task_reference] = task
        else:
            tasks_to_run_filtered[task_reference] = task

        for a in [
            task_reference_constants.MANIFEST_SECTION_NAMES,
            task_reference_constants.MANIFEST_ITEM_NAMES,
            task_reference_constants.MANIFEST_ACCOUNT_IDS,
        ]:
            if task.get(a):
                del task[a]

    executor_account_id = config.get_executor_account_id()
    is_dry_run = is_list_launches = False
    execution_mode = config.get_execution_mode()
    on_complete_url = config.get_on_complete_url()
    running_exploded = False
    output_cache_starting_point = ""

    runner.run_tasks(
        puppet_account_id,
        executor_account_id,
        tasks_to_run,
        num_workers,
        is_dry_run,
        is_list_launches,
        execution_mode,
        on_complete_url,
        running_exploded,
        tasks_to_run_filtered,
        path,
        f,
    )



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
from servicecatalog_puppet import constants
from servicecatalog_puppet.commands.task_reference_helpers.generators import (
    c7n_aws_lambdas,
    imported_portfolios,
    launches,
    organizational_units,
    service_control_policies,
    spoke_local_portfolios,
    stacks,
    tag_policies,
    workspaces,
)


def generate(
    all_tasks,
    all_tasks_task_reference,
    item_name,
    puppet_account_id,
    section_name,
    task_reference,
    task_to_add,
    manifest,
):

    if section_name == constants.LAUNCHES:
        launches.handle_launches(
            all_tasks,
            all_tasks_task_reference,
            item_name,
            puppet_account_id,
            section_name,
            task_reference,
            task_to_add,
        )

    if section_name == constants.SPOKE_LOCAL_PORTFOLIOS:
        spoke_local_portfolios.handle_spoke_local_portfolios(
            all_tasks,
            all_tasks_task_reference,
            item_name,
            puppet_account_id,
            section_name,
            task_reference,
            task_to_add,
        )

    if section_name == constants.IMPORTED_PORTFOLIOS:
        imported_portfolios.handle_imported_portfolios(
            all_tasks,
            all_tasks_task_reference,
            item_name,
            puppet_account_id,
            section_name,
            task_reference,
            task_to_add,
        )

    if section_name == constants.WORKSPACES:
        workspaces.handle_workspaces(
            all_tasks,
            all_tasks_task_reference,
            item_name,
            puppet_account_id,
            section_name,
            task_reference,
            task_to_add,
        )

    if section_name == constants.STACKS:
        stacks.handle_stacks(
            all_tasks,
            all_tasks_task_reference,
            item_name,
            puppet_account_id,
            section_name,
            task_reference,
            task_to_add,
        )

    if section_name == constants.SERVICE_CONTROL_POLICIES:
        service_control_policies.handle_service_control_policies(
            all_tasks,
            all_tasks_task_reference,
            item_name,
            puppet_account_id,
            section_name,
            task_reference,
            task_to_add,
        )

    if section_name == constants.TAG_POLICIES:
        tag_policies.handle_tag_policies(
            all_tasks,
            all_tasks_task_reference,
            item_name,
            puppet_account_id,
            section_name,
            task_reference,
            task_to_add,
        )

    if section_name == constants.ORGANIZATIONAL_UNITS:
        organizational_units.handle_organizational_units(
            all_tasks,
            all_tasks_task_reference,
            item_name,
            puppet_account_id,
            section_name,
            task_reference,
            task_to_add,
        )

    if section_name == constants.C7N_AWS_LAMBDAS:
        c7n_aws_lambdas.handle_c7n_aws_lambdas(
            all_tasks,
            all_tasks_task_reference,
            item_name,
            puppet_account_id,
            section_name,
            task_reference,
            task_to_add,
            manifest,
        )



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks


retry_max_attempts = 50


class GetAllProductsAndTheirVersionsTask(tasks.TaskWithReference):
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    portfolio = luigi.Parameter()
    portfolio_task_reference = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
        }

    def run(self):
        portfolio_details = self.get_output_from_reference_dependency(
            self.portfolio_task_reference
        )
        portfolio_id = portfolio_details.get("Id")
        if portfolio_id is None:
            self.write_empty_output()
        else:
            products = dict()
            with self.spoke_regional_client(
                "servicecatalog", retry_max_attempts=retry_max_attempts
            ) as servicecatalog:
                paginator = servicecatalog.get_paginator("search_products_as_admin")
                for page in paginator.paginate(
                    PortfolioId=portfolio_id,
                    ProductSource="ACCOUNT",
                ):
                    for product_view_detail in page.get("ProductViewDetails", []):
                        product_arn = product_view_detail.get("ProductARN")
                        product_view_summary = product_view_detail.get(
                            "ProductViewSummary"
                        )
                        product_view_summary["ProductArn"] = product_arn
                        product_view_summary["Versions"] = dict()
                        product_id = product_view_summary.get("ProductId")
                        product_name = product_view_summary.get("Name")
                        products[product_name] = product_view_summary
                        provisioning_artifact_summaries = (
                            servicecatalog.describe_product_as_admin(
                                Id=product_view_summary.get("ProductId"),
                            ).get("ProvisioningArtifactSummaries")
                        )
                        for (
                            provisioning_artifact_summary
                        ) in provisioning_artifact_summaries:
                            version_name = provisioning_artifact_summary.get("Name")
                            provisioning_artifact_detail = (
                                servicecatalog.describe_provisioning_artifact(
                                    ProductId=product_id,
                                    ProvisioningArtifactName=version_name,
                                ).get("ProvisioningArtifactDetail")
                            )
                            provisioning_artifact_summary["Active"] = (
                                provisioning_artifact_detail.get("Active")
                            )
                            provisioning_artifact_summary["Guidance"] = (
                                provisioning_artifact_detail.get("Guidance")
                            )
                            product_view_summary["Versions"][
                                version_name
                            ] = provisioning_artifact_summary
            self.write_output(products)



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import time

import luigi

from servicecatalog_puppet import constants, serialisation_utils
from servicecatalog_puppet.workflow.dependencies import tasks


class DoTerminateProductTask(tasks.TaskWithReference):
    launch_name = luigi.Parameter()

    region = luigi.Parameter()
    account_id = luigi.Parameter()

    portfolio = luigi.Parameter()
    product = luigi.Parameter()
    version = luigi.Parameter()

    ssm_param_inputs = luigi.ListParameter(default=[], significant=False)

    launch_parameters = luigi.DictParameter(default={}, significant=False)
    manifest_parameters = luigi.DictParameter(default={}, significant=False)
    account_parameters = luigi.DictParameter(default={}, significant=False)

    retry_count = luigi.IntParameter(default=1, significant=False)
    worker_timeout = luigi.IntParameter(default=0, significant=False)
    ssm_param_outputs = luigi.ListParameter(default=[], significant=False)
    requested_priority = luigi.IntParameter(significant=False, default=0)

    execution = luigi.Parameter()

    try_count = 1
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "launch_name": self.launch_name,
            "account_id": self.account_id,
            "region": self.region,
        }

    def run(self):
        with self.spoke_regional_client("servicecatalog") as service_catalog:
            try:
                service_catalog.describe_provisioned_product(Name=self.launch_name)
            except service_catalog.exceptions.ResourceNotFoundException:
                self.write_empty_output()
                return

            record_detail = service_catalog.terminate_provisioned_product(
                ProvisionedProductName=self.launch_name
            ).get("RecordDetail")
            record_id = record_detail.get("RecordId")

            status = "IN_PROGRESS"
            while status == "IN_PROGRESS":
                record_detail = service_catalog.describe_record(Id=record_id).get(
                    "RecordDetail"
                )
                status = record_detail.get("Status")
                self.info(f"termination of {self.launch_name} current status: {status}")
                if status != "IN_PROGRESS":
                    break
                else:
                    time.sleep(3)

            if status not in ["CREATED", "SUCCEEDED"]:
                self.info(serialisation_utils.dump(record_detail.get("RecordErrors")))
                raise Exception(
                    f"Failed to terminate provisioned product: Status = {status}"
                )

            self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import os

import luigi

from servicecatalog_puppet import (
    constants,
    serialisation_utils,
    task_reference_constants,
)
from servicecatalog_puppet.workflow.dependencies import tasks


class RunDeployInSpokeTask(tasks.TaskWithReference):
    account_id = luigi.Parameter()
    generate_manifest_ref = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_NO_CACHE

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
            "puppet_account_id": self.puppet_account_id,
            "account_id": self.account_id,
        }

    def run(self):
        generated_manifest = self.get_output_from_reference_dependency(
            self.generate_manifest_ref
        )
        build_spec = generated_manifest.get("build_spec")
        spoke_execution_mode_deploy_env = generated_manifest.get(
            "spoke_execution_mode_deploy_env"
        )
        vars = generated_manifest.get("vars")

        bucket = f"sc-puppet-spoke-deploy-{self.puppet_account_id}"
        with self.hub_client("s3") as s3:
            task_reference_content = open(
                self.manifest_task_reference_file_path.replace(
                    ".json", f"-{self.account_id}.json"
                ),
                "r",
            ).read()
            key = f"{os.getenv('CODEBUILD_BUILD_NUMBER', '0')}-reference-for-{self.account_id}.json"
            self.debug(f"Uploading task reference {key} to {bucket}")
            s3.put_object(
                Body=task_reference_content,
                Bucket=bucket,
                Key=key,
            )
            self.debug(f"Generating presigned URL for {key}")
            reference_signed_url = s3.generate_presigned_url(
                "get_object",
                Params={"Bucket": bucket, "Key": key},
                ExpiresIn=60 * 60 * 24,
            )
            vars.append(
                {
                    "name": "TASK_REFERENCE_URL",
                    "value": reference_signed_url,
                    "type": "PLAINTEXT",
                },
            )

        with self.spoke_client("codebuild") as codebuild:
            response = codebuild.start_build(
                projectName=constants.EXECUTION_SPOKE_CODEBUILD_PROJECT_NAME,
                environmentVariablesOverride=vars,
                computeTypeOverride=spoke_execution_mode_deploy_env,
                buildspecOverride=build_spec,
                imageOverride=constants.CODEBUILD_DEFAULT_IMAGE,
            )
        self.write_output(dict(account_id=self.account_id, **response))



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import os
from unittest import mock as mocker
from unittest.mock import call


def test_resolve_from_site_packages():
    # setup
    from servicecatalog_puppet import asset_helpers as sut

    expected_result = os.path.sep.join(
        [os.path.dirname(os.path.abspath(__file__)), "foo"]
    )

    # exercise
    actual_result = sut.resolve_from_site_packages("foo")

    # verify
    assert actual_result == expected_result


@mocker.patch("builtins.open", new_callable=mocker.MagicMock())
def test_read_from_site_packages(mocked_open):
    # setup
    from servicecatalog_puppet import asset_helpers as sut

    expected_param = os.path.sep.join(
        [os.path.dirname(os.path.abspath(__file__)), "foo"]
    )

    # exercise
    sut.read_from_site_packages("foo")

    # verify
    assert mocked_open.call_count == 1
    assert mocked_open.call_args == call(expected_param, "r")



#   Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#   SPDX-License-Identifier: Apache-2.0
from servicecatalog_puppet import utils


def generate_stack_name_for_associations_by_item_name(item_name):
    return (
        f"associations-v2-for-{utils.slugify_for_cloudformation_stack_name(item_name)}"
    )



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import time

import networkx as nx

from servicecatalog_puppet.waluigi.constants import (
    CONTROL_EVENT__COMPLETE,
    QUEUE_STATUS,
)
from servicecatalog_puppet.waluigi.dag_utils import build_the_dag, logger


def scheduler_task(
    num_workers,
    task_queue,
    results_queue,
    control_queue,
    control_event,
    queue_refill_sleep_duration,
    all_tasks,
    tasks_to_run,
):
    number_of_target_tasks_in_flight = num_workers
    should_be_running = True
    if control_event:
        should_be_running = not control_event.is_set()
    while should_be_running:
        dag = build_the_dag(all_tasks)
        generations = list(nx.topological_generations(dag))
        if not generations:
            logger.info("No more batches to run")
            if control_queue:
                control_queue.put(CONTROL_EVENT__COMPLETE)
            if control_event:
                control_event.set()
            return

        current_generation = list(generations[-1])
        number_of_tasks_in_flight = 0
        number_of_tasks_processed = 0
        number_of_tasks_in_generation = len(current_generation)
        current_generation_in_progress = True

        while current_generation_in_progress:
            logger.info("starting batch")
            # start each iteration by checking if the queue has enough jobs in it
            while (
                current_generation
                and number_of_tasks_in_flight < number_of_target_tasks_in_flight
            ):
                # there are enough jobs in the queue
                number_of_tasks_in_flight += 1
                task_to_run_reference = current_generation.pop()
                logger.info(f"sending: {task_to_run_reference}")
                task_queue.put(task_to_run_reference)
                time.sleep(queue_refill_sleep_duration)

            # now handle a complete jobs from the workers
            task_reference, result = results_queue.get()
            if task_reference:
                number_of_tasks_in_flight -= 1
                number_of_tasks_processed += 1
                logger.info(
                    f"receiving: [{number_of_tasks_processed}]: {task_reference}, {result}"
                )
                task_just_run = all_tasks[task_reference]
                task_just_run[QUEUE_STATUS] = result
                all_tasks[task_reference] = task_just_run

            if not current_generation:  # queue now empty - wait for all to complete
                logger.info("tasks now scheduled")
                while number_of_tasks_processed < number_of_tasks_in_generation:
                    task_reference, result = results_queue.get()
                    if task_reference:
                        number_of_tasks_in_flight -= 1
                        number_of_tasks_processed += 1
                        logger.info(
                            f"receiving: [{number_of_tasks_processed}]: {task_reference}, {result}"
                        )
                        task_just_run = all_tasks[task_reference]
                        task_just_run[QUEUE_STATUS] = result
                        all_tasks[task_reference] = task_just_run
                else:
                    current_generation_in_progress = False
                    logger.info("finished batch")
        if control_event:
            should_be_running = not control_event.is_set()
    logger.info("finished all batches")



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import json
import time

import luigi

import servicecatalog_puppet.manifest_utils
import servicecatalog_puppet.serialisation_utils
from servicecatalog_puppet import aws, constants
from servicecatalog_puppet.workflow.dependencies import tasks


class ProvisionProductTask(tasks.TaskWithParameters):
    manifest_file_path = luigi.Parameter()
    launch_name = luigi.Parameter()

    region = luigi.Parameter()
    account_id = luigi.Parameter()

    portfolio = luigi.Parameter()
    product = luigi.Parameter()
    version = luigi.Parameter()

    portfolio_get_all_products_and_their_versions_ref = luigi.Parameter()
    describe_provisioning_params_ref = luigi.Parameter()

    ssm_param_inputs = luigi.ListParameter(default=[], significant=False)

    launch_parameters = luigi.DictParameter(default={}, significant=False)
    manifest_parameters = luigi.DictParameter(default={}, significant=False)
    account_parameters = luigi.DictParameter(default={}, significant=False)

    retry_count = luigi.IntParameter(default=1, significant=False)
    worker_timeout = luigi.IntParameter(default=0, significant=False)
    ssm_param_outputs = luigi.ListParameter(default=[], significant=False)
    requested_priority = luigi.IntParameter(significant=False, default=0)

    execution = luigi.Parameter()
    tags = luigi.ListParameter()

    section_name = constants.LAUNCHES
    cachable_level = constants.CACHE_LEVEL_RUN
    need_to_provision = False

    @property
    def item_name(self):
        return self.launch_name

    try_count = 1

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "launch_name": self.launch_name,
            "account_id": self.account_id,
            "region": self.region,
            "need_to_provision": self.need_to_provision,
        }

    @property
    def priority(self):
        return self.requested_priority

    def run(self):
        products_and_their_versions = self.get_output_from_reference_dependency(
            self.portfolio_get_all_products_and_their_versions_ref
        )
        product = products_and_their_versions.get(self.product)
        product_id = product.get("ProductId")
        version_id = product.get("Versions").get(self.version).get("Id")
        describe_provisioning_params = self.get_output_from_reference_dependency(
            self.describe_provisioning_params_ref
        )

        task_output = dict(
            **self.params_for_results_display(),
            account_parameters=servicecatalog_puppet.serialisation_utils.unwrap(
                self.account_parameters
            ),
            launch_parameters=servicecatalog_puppet.serialisation_utils.unwrap(
                self.launch_parameters
            ),
            manifest_parameters=servicecatalog_puppet.serialisation_utils.unwrap(
                self.manifest_parameters
            ),
        )

        all_params = self.get_parameter_values()

        with self.spoke_regional_client("servicecatalog") as service_catalog:
            with self.spoke_regional_client("cloudformation") as cloudformation:
                path_name = self.portfolio
                # Get the status
                (
                    provisioned_product_detail,
                    was_provisioned_before,
                ) = self.check_was_previously_provisioned(service_catalog)

                params_to_use = {}
                provisioned_product_id = False
                # provisioning_artifact_parameters = self.load_from_input(
                #     "provisioning_artifact_parameters"
                # )
                for p in describe_provisioning_params:
                    param_name = p.get("ParameterKey")
                    params_to_use[param_name] = all_params.get(
                        param_name, p.get("DefaultValue")
                    )

                if was_provisioned_before:
                    (
                        ignore_me,
                        provisioning_artifact_id,
                        provisioned_product_id,
                    ) = self.clean_up_existing_provisioned_product(
                        provisioned_product_detail, service_catalog
                    )

                    self.need_to_provision = True

                    if provisioning_artifact_id == version_id:
                        self.info(f"found previous good provision")
                        if provisioned_product_id:
                            self.info(f"checking params for diffs")
                            pp_stack_name = aws.get_stack_name_for_pp_id(
                                service_catalog, provisioned_product_id
                            )
                            with self.spoke_regional_client(
                                "cloudformation"
                            ) as cloudformation:
                                provisioned_parameters = aws.get_parameters_for_stack(
                                    cloudformation,
                                    pp_stack_name,
                                )
                            self.info(f"current params: {provisioned_parameters}")
                            self.info(f"new params: {params_to_use}")

                            if provisioned_parameters == params_to_use:
                                self.info(f"params unchanged")
                                self.need_to_provision = False
                            else:
                                self.info(f"params changed")
                else:
                    self.need_to_provision = True

                task_output["provisioned"] = self.need_to_provision
                task_output["section_name"] = self.section_name
                if self.need_to_provision:
                    self.info(
                        f"about to provision with params: {json.dumps(servicecatalog_puppet.serialisation_utils.unwrap(params_to_use))}"
                    )

                    if provisioned_product_id:
                        pp_stack_name = aws.get_stack_name_for_pp_id(
                            service_catalog, provisioned_product_id
                        )
                        stack = aws.get_stack_output_for(
                            cloudformation,
                            pp_stack_name,
                        )
                        stack_status = stack.get("StackStatus")
                        self.info(f"current cfn stack_status is {stack_status}")
                        if stack_status not in [
                            "UPDATE_COMPLETE",
                            "CREATE_COMPLETE",
                            "UPDATE_ROLLBACK_COMPLETE",
                        ]:
                            raise Exception(
                                f"[{self.task_reference}] current cfn stack_status is {stack_status}"
                            )
                        if stack_status == "UPDATE_ROLLBACK_COMPLETE":
                            self.warning(
                                f"{pp_stack_name} has a status of "
                                f"{stack_status}.  This may need manual resolution."
                            )

                    if provisioned_product_id:
                        if self.should_use_product_plans:
                            path_id = aws.get_path_for_product(
                                service_catalog, product_id, self.portfolio
                            )
                            provisioned_product_id = aws.provision_product_with_plan(
                                service_catalog,
                                self.launch_name,
                                self.account_id,
                                self.region,
                                product_id,
                                version_id,
                                self.puppet_account_id,
                                path_id,
                                params_to_use,
                                self.version,
                                self.should_use_sns,
                                self.tags,
                            )
                        else:
                            provisioned_product_id = aws.update_provisioned_product(
                                service_catalog,
                                self.launch_name,
                                self.account_id,
                                self.region,
                                product_id,
                                version_id,
                                self.puppet_account_id,
                                path_name,
                                params_to_use,
                                self.version,
                                self.execution,
                                self.tags,
                            )

                    else:
                        provisioned_product_id = aws.provision_product(
                            service_catalog,
                            self.launch_name,
                            self.account_id,
                            self.region,
                            product_id,
                            version_id,
                            self.puppet_account_id,
                            path_name,
                            params_to_use,
                            self.version,
                            self.should_use_sns,
                            self.execution,
                            self.tags,
                        )

        self.write_output(task_output)
        self.info("finished")

    def clean_up_existing_provisioned_product(
        self, provisioned_product_detail, service_catalog
    ):
        provisioned_product_id = provisioned_product_detail.get("Id")
        provisioning_artifact_id = provisioned_product_detail.get(
            "ProvisioningArtifactId"
        )
        product_id = provisioned_product_detail.get("ProductId")
        # Wait for it to complete what it is doing
        while provisioned_product_detail.get("Status") in [
            "UNDER_CHANGE",
            "PLAN_IN_PROGRESS",
        ]:
            time.sleep(1)
            provisioned_product_detail = service_catalog.describe_provisioned_product(
                Name=self.launch_name
            ).get("ProvisionedProductDetail")
        # Delete it if it is non usable
        if provisioned_product_detail.get("Status") in [
            "ERROR",
        ]:
            record_detail = service_catalog.terminate_provisioned_product(
                ProvisionedProductName=self.launch_name
            ).get("RecordDetail")
            while record_detail.get("Status") in [
                "IN_PROGRESS",
                "IN_PROGRESS_IN_ERROR",
            ]:
                time.sleep(1)
                service_catalog.describe_record(Id=record_detail.get("RecordId"))

            if record_detail.get("Status") in [
                "CREATED",
                "FAILED",
            ]:
                raise Exception(
                    f"Unexpected record detail status: {record_detail.get('Status')}"
                )
            elif record_detail.get("Status") in [
                "SUCCEEDED",
            ]:
                pass
            else:
                raise Exception(
                    f"Unhandled record detail status: {record_detail.get('Status')}"
                )

        elif provisioned_product_detail.get("Status") in [
            "TAINTED",
        ]:
            self.warning(f"Provisioned Product is tainted")
        return product_id, provisioning_artifact_id, provisioned_product_id

    def check_was_previously_provisioned(self, service_catalog):
        was_provisioned_before = True
        try:
            provisioned_product_detail = service_catalog.describe_provisioned_product(
                Name=self.launch_name
            ).get("ProvisionedProductDetail")
        except service_catalog.exceptions.ResourceNotFoundException:
            was_provisioned_before = False
            provisioned_product_detail = None
        return provisioned_product_detail, was_provisioned_before



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import unittest

from servicecatalog_puppet import constants, task_reference_constants


class Boto3ParameterHandlerTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        from servicecatalog_puppet.commands.task_reference_helpers.generators import (
            boto3_parameter_handler,
        )

        self.sut = boto3_parameter_handler

    def test_for_boto3_parameters(self):
        # setup
        puppet_account_id = "hub_account_id"
        account_id = "spoke_account_id"
        region = "eu-west-0"
        item_name = "depsrefactor"
        dependency_item_name = "something_else"

        section_name = constants.STACKS
        task = {
            "stack_name": item_name,
            "launch_name": "",
            "stack_set_name": "",
            "capabilities": ["CAPABILITY_NAMED_IAM"],
            "ssm_param_outputs": [],
            "bucket": f"sc-puppet-stacks-repository-{puppet_account_id}",
            "key": f"stack/{item_name}/v1/stack.template-${{AWS::Region}}.yaml",
            "version_id": "",
            "execution": "spoke",
            "use_service_role": True,
            "tags": [],
            "puppet_account_id": puppet_account_id,
            "status": None,
            "requested_priority": 0,
            "dependencies": [
                {
                    "affinity": "stack",
                    "name": dependency_item_name,
                    "type": "stack",
                }
            ],
            "account_id": account_id,
            "region": region,
            task_reference_constants.MANIFEST_SECTION_NAMES: {"stacks": True},
            task_reference_constants.MANIFEST_ITEM_NAMES: {item_name: True},
            task_reference_constants.MANIFEST_ACCOUNT_IDS: {account_id: True},
            "section_name": "stacks",
            "item_name": item_name,
            "dependencies_by_reference": [
                "create-policies",
                f"prepare-account-for-stacks-{account_id}",
                f"get-template-from-s3-stacks-{item_name}",
            ],
            "task_reference": f"stacks_ccoe-{item_name}_{account_id}_{region}",
            "get_s3_template_ref": f"get-template-from-s3-stacks-{item_name}",
        }
        task_reference = task["task_reference"]
        new_tasks = {task_reference: task}

        describe_stacks_filter = "xyz"

        output_key = "SomeOutputKey"

        parameter_details = {
            "boto3": {
                "account_id": "${AWS::AccountId}",
                "arguments": {"StackName": item_name},
                "call": "describe_stacks",
                "client": "cloudformation",
                "filter": describe_stacks_filter,
                "region": "${AWS::Region}",
                "use_paginator": True,
            },
            "cloudformation_stack_output": {
                "account_id": "${AWS::AccountId}",
                "output_key": output_key,
                "region": "${AWS::Region}",
                "stack_name": dependency_item_name,
            },
        }

        parameter_name = "some_parameter"

        # exercise
        self.sut.boto3_parameter_handler(
            new_tasks,
            parameter_details,
            parameter_name,
            puppet_account_id,
            task,
        )

        # verify
        n_new_tasks = len(new_tasks.keys())
        boto3_parameter_task_reference = f"{constants.BOTO3_PARAMETERS}-cloudformation_stack_output-{dependency_item_name}-{output_key}-{account_id}-{region}"
        self.assertEqual(
            [
                "create-policies",
                f"prepare-account-for-stacks-{account_id}",
                f"get-template-from-s3-stacks-{item_name}",
                boto3_parameter_task_reference,
            ],
            new_tasks[task_reference].get("dependencies_by_reference"),
            "assert new dependency is added to the task needing the parameter",
        )

        expected_boto3_task_ref = f"{constants.BOTO3_PARAMETERS}-{section_name}-{item_name}-{parameter_name}-{account_id}-{region}"
        self.assertEqual(
            {
                "status": None,
                "execution": "spoke",
                "task_reference": boto3_parameter_task_reference,
                "dependencies_by_reference": [],
                "dependencies": [],
                task_reference_constants.MANIFEST_SECTION_NAMES: {
                    constants.STACKS: True
                },
                task_reference_constants.MANIFEST_ITEM_NAMES: {item_name: True},
                task_reference_constants.MANIFEST_ACCOUNT_IDS: {account_id: True},
                "account_id": account_id,
                "region": region,
                "arguments": {"StackName": item_name},
                "call": "describe_stacks",
                "client": "cloudformation",
                "filter": describe_stacks_filter,
                "use_paginator": True,
                "section_name": constants.BOTO3_PARAMETERS,
            },
            new_tasks[boto3_parameter_task_reference],
            "assert the boto3 task is generated correctly",
        )

        self.assertEqual(2, n_new_tasks, "assert the correct number of tasks exist")



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks


class GetOrCreateOrganizationalUnitTask(tasks.TaskWithReference):
    region = luigi.Parameter()
    account_id = luigi.Parameter()

    path = luigi.Parameter()
    parent_ou_id = luigi.Parameter()
    name = luigi.Parameter()
    tags = luigi.ListParameter()
    parent_ou_task_ref = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "region": self.region,
            "path": self.path,
            "parent_ou_id": self.parent_ou_id,
            "name": self.name,
            "task_reference": self.task_reference,
        }

    def find_ou_in_parent(self, orgs, parent_id):
        paginator = orgs.get_paginator("list_children")
        for page in paginator.paginate(
            ParentId=parent_id, ChildType="ORGANIZATIONAL_UNIT"
        ):
            for child in page.get("Children", []):
                child_details = orgs.describe_organizational_unit(
                    OrganizationalUnitId=child.get("Id")
                ).get("OrganizationalUnit")
                if child_details.get("Name") == self.name:
                    return child_details.get("Id")
        return None

    def run(self):
        with self.organizations_policy_client() as orgs:
            if self.path == "/":
                result = [orgs.convert_path_to_ou(self.path)]
            else:
                result = self.get_output_from_reference_dependency(
                    self.parent_ou_task_ref
                )
                parent_id = result[-1]
                ou_id = self.find_ou_in_parent(orgs, parent_id)
                if ou_id:
                    result.append(ou_id)
                else:
                    organizational_unit = orgs.create_organizational_unit(
                        ParentId=parent_id,
                        Name=self.name,
                        Tags=[
                            dict(Key=tag.get("Key"), Value=tag.get("Value"))
                            for tag in self.tags
                        ],
                    ).get("OrganizationalUnit")
                    result.append(organizational_unit.get("Id"))

        self.write_output(result)



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
from servicecatalog_puppet import constants, task_reference_constants


def handle_stacks(
    all_tasks,
    all_tasks_task_reference,
    item_name,
    puppet_account_id,
    section_name,
    task_reference,
    task_to_add,
):
    prepare_task_reference = (
        f"{constants.PREPARE_ACCOUNT_FOR_STACKS}-{task_to_add.get('account_id')}"
    )
    if not all_tasks.get(prepare_task_reference):
        all_tasks[prepare_task_reference] = {
            "task_reference": prepare_task_reference,
            "account_id": task_to_add.get("account_id"),
            "region": constants.HOME_REGION,
            "puppet_account_id": puppet_account_id,
            "dependencies_by_reference": [constants.CREATE_POLICIES],
            task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
            task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
            task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
            "section_name": constants.PREPARE_ACCOUNT_FOR_STACKS,
        }
    all_tasks[all_tasks_task_reference]["dependencies_by_reference"].append(
        prepare_task_reference
    )
    all_tasks[prepare_task_reference][
        task_reference_constants.MANIFEST_SECTION_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
    all_tasks[prepare_task_reference][
        task_reference_constants.MANIFEST_ITEM_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
    all_tasks[prepare_task_reference][
        task_reference_constants.MANIFEST_ACCOUNT_IDS
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))

    get_s3_template_ref = f"{constants.GET_TEMPLATE_FROM_S3}-{section_name}-{item_name}"
    if not all_tasks.get(get_s3_template_ref):
        all_tasks[get_s3_template_ref] = {
            "task_reference": get_s3_template_ref,
            "execution": "hub",
            "bucket": task_to_add.get("bucket"),
            "key": task_to_add.get("key"),
            "region": task_to_add.get("region"),
            "version_id": task_to_add.get("version_id"),
            "puppet_account_id": puppet_account_id,
            "account_id": puppet_account_id,
            "dependencies_by_reference": [constants.CREATE_POLICIES],
            task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
            task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
            task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
            "section_name": constants.GET_TEMPLATE_FROM_S3,
        }
    all_tasks[all_tasks_task_reference]["get_s3_template_ref"] = get_s3_template_ref
    all_tasks[all_tasks_task_reference]["dependencies_by_reference"].append(
        get_s3_template_ref
    )
    all_tasks[get_s3_template_ref][
        task_reference_constants.MANIFEST_SECTION_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
    all_tasks[get_s3_template_ref][task_reference_constants.MANIFEST_ITEM_NAMES].update(
        task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES)
    )
    all_tasks[get_s3_template_ref][
        task_reference_constants.MANIFEST_ACCOUNT_IDS
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks


class ImportIntoSpokeLocalPortfolioTask(tasks.TaskWithReference):
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    portfolio_task_reference = luigi.Parameter()
    hub_portfolio_task_reference = luigi.Parameter()

    portfolio_get_all_products_and_their_versions_ref = luigi.Parameter()
    portfolio_get_all_products_and_their_versions_for_hub_ref = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
            "puppet_account_id": self.puppet_account_id,
            "region": self.region,
            "account_id": self.account_id,
        }

    def run(self):
        spoke_portfolio_details = self.get_output_from_reference_dependency(
            self.portfolio_task_reference
        )
        spoke_portfolio_id = spoke_portfolio_details.get("Id")
        spoke_products_and_their_versions = self.get_output_from_reference_dependency(
            self.portfolio_get_all_products_and_their_versions_ref
        )
        hub_portfolio_details = self.get_output_from_reference_dependency(
            self.hub_portfolio_task_reference
        )
        hub_portfolio_id = hub_portfolio_details.get("Id")
        hub_products_and_their_versions = self.get_output_from_reference_dependency(
            self.portfolio_get_all_products_and_their_versions_for_hub_ref
        )

        with self.spoke_regional_client("servicecatalog") as servicecatalog:
            products_to_check = list()
            for (
                hub_product_name,
                hub_product_details,
            ) in hub_products_and_their_versions.items():
                if spoke_products_and_their_versions.get(hub_product_name) is None:
                    product_id = hub_product_details.get("ProductId")
                    self.info(f"Associating {product_id}")
                    servicecatalog.associate_product_with_portfolio(
                        ProductId=product_id,
                        PortfolioId=spoke_portfolio_id,
                        SourcePortfolioId=hub_portfolio_id,
                    )
                    products_to_check.append(product_id)

            n_products_to_check = len(products_to_check)
            products_found = 0
            while products_found < n_products_to_check:
                products_ids = list()
                paginator = servicecatalog.get_paginator("search_products_as_admin")
                for page in paginator.paginate(
                    PortfolioId=spoke_portfolio_id,
                ):
                    for pvd in page.get("ProductViewDetails", []):
                        products_ids.append(
                            pvd.get("ProductViewSummary").get("ProductId")
                        )

                products_found = 0
                for product_to_check in products_to_check:
                    if product_to_check in products_ids:
                        products_found += 1
            self.info("Finished waiting for association of products to portfolio")
            self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import unittest

from servicecatalog_puppet.waluigi.constants import RESOURCES_REQUIRED


class TestExternal(unittest.TestCase):
    def setUp(self) -> None:
        self.maxDiff = None
        from servicecatalog_puppet.waluigi.locks import external

        self.sut = external

    def test_are_resources_are_free_for_task_dict_with_no_locked_resources(self):
        # setup
        resources_in_use = dict()
        task_parameters = dict()
        task_parameters[RESOURCES_REQUIRED] = ["aaaa"]
        expected_result = True

        # exercise
        actual_result, _ = self.sut.are_resources_are_free_for_task_dict(
            task_parameters, resources_in_use
        )

        # verify
        self.assertEqual(expected_result, actual_result)

    def test_are_resources_are_free_for_task_dict_with_other_locked_resources(self):
        # setup
        resources_in_use = dict(bbbb="some-other-task")
        task_parameters = dict()
        task_parameters[RESOURCES_REQUIRED] = ["aaaa"]
        expected_result = True

        # exercise
        actual_result, _ = self.sut.are_resources_are_free_for_task_dict(
            task_parameters, resources_in_use
        )

        # verify
        self.assertEqual(expected_result, actual_result)

    def test_are_resources_are_free_for_task_dict_with_locked_resources(self):
        # setup
        resources_in_use = dict(aaaa="some-other-task")
        task_parameters = dict()
        task_parameters[RESOURCES_REQUIRED] = ["aaaa"]
        expected_result = False

        # exercise
        actual_result, _ = self.sut.are_resources_are_free_for_task_dict(
            task_parameters, resources_in_use
        )

        # verify
        self.assertEqual(expected_result, actual_result)

    def test_are_resources_are_free_for_task_dict_with_locked_resources_and_more(self):
        # setup
        resources_in_use = dict(aaaa="some-other-task", bbb="sdsnkdnsk")
        task_parameters = dict()
        task_parameters[RESOURCES_REQUIRED] = ["aaaa"]
        expected_result = False

        # exercise
        actual_result, _ = self.sut.are_resources_are_free_for_task_dict(
            task_parameters, resources_in_use
        )

        # verify
        self.assertEqual(expected_result, actual_result)



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet import serialisation_utils
from servicecatalog_puppet.waluigi.constants import RESOURCES_REQUIRED
from servicecatalog_puppet.waluigi.dag_utils import logger


def are_resources_are_free_for_task(task_parameters: dict, resources_file_path: str):
    with open(resources_file_path, "rb") as f:
        resources_in_use = serialisation_utils.json_loads(f.read())
    return are_resources_are_free_for_task_dict(task_parameters, resources_in_use)


def are_resources_are_free_for_task_dict(task_parameters, resources_in_use):
    return (
        all(
            resources_in_use.get(r, False) is False
            for r in task_parameters.get(RESOURCES_REQUIRED, [])
        ),
        resources_in_use,
    )


def lock_resources_for_task(
    task_reference: str,
    task_parameters: dict,
    resources_in_use: dict,
    resources_file_path: str,
):
    # print(f"Worker locking {task_reference}")
    for r in task_parameters.get(RESOURCES_REQUIRED, []):
        resources_in_use[r] = task_reference
    with open(resources_file_path, "wb") as f:
        f.write(serialisation_utils.json_dumps(resources_in_use))


def unlock_resources_for_task(task_parameters: dict, resources_file_path: str):
    with open(resources_file_path, "rb") as f:
        resources_in_use = serialisation_utils.json_loads(f.read())
    for r in task_parameters.get(RESOURCES_REQUIRED, []):
        try:
            del resources_in_use[r]
        except KeyError:
            logger.warn(
                f"{task_parameters.get('task_reference')} tried to unlock {r} but it wasn't present"
            )
    with open(resources_file_path, "wb") as f:
        f.write(serialisation_utils.json_dumps(resources_in_use))
